import obspy
from obspy import UTCDateTime
from scipy import interpolate
import numpy as np
import os
from seisloc.utils import month_day

def aligned_sac_datas(tr1,tr2,scc_dt,tb,te,marker='t0'):
    """
    Parameters:
       tr1,tr2: obspy trace
        scc_dt: time difference measured by cross-correlation, t1-t2
            tb: Waveform cut time before the S arrival
            te: Waveform cut time after the S arrival
        marker: 'a' for P arrival time, 't0' for S arrival time
    """
    tr1.detrend("linear")
    tr1.detrend("constant")
    tr2.detrend("linear")
    tr2.detrend("constant")
    
    t01 = getattr(tr1.stats.sac,marker)
    t02 = getattr(tr2.stats.sac,marker)
    b1 = tr1.stats.sac.b
    b2 = tr2.stats.sac.b
    trim_b1 = tr1.stats.starttime+(t01-b1)+scc_dt+tb
    trim_e1 = trim_b1+(te-tb)

    trim_b2 = tr2.stats.starttime+(t01-b2)+tb
    trim_e2 = trim_b2+(te-tb)

    tr1.trim(starttime = trim_b1, endtime = trim_e1)
    tr2.trim(starttime = trim_b2, endtime = trim_e2)
    
    tr1.detrend("constant")
    tr1.detrend("linear")
    tr2.detrend("constant")
    tr2.detrend("linear")

    return tr1.data,tr2.data

def sac_interpolation(insacPth,outsacPth,factor=10):
    """
    Interpolate waveform data using the quadratic fitting method provided by scipy.interpolate.interp1d module
    """
    insacPth = os.path.abspath(insacPth)
    st = obspy.read(insacPth)
    data = st[0].data
    delta = st[0].stats.delta
    xs = np.arange(len(data))
    f = interpolate.interp1d(xs,data,kind='quadratic')
    
    deltaNew = delta/factor
    ixs = np.arange((len(data)-1)*factor+1)*1/factor
    dataInterp = f(ixs)
    stInterp = st.copy()
    stInterp[0].data = dataInterp
    stInterp[0].stats.delta = deltaNew
    stInterp[0].write(outsacPth,format="SAC")
    
def read_sac_ref_time(tr):
    """
    Read and return reference time of a sac file in obspy.UTCDateTime format.

    Parameter
    --------
    tr: Trace object of obspy
    """

    nzyear = tr.stats.sac.nzyear
    nzjday = tr.stats.sac.nzjday
    nzhour = tr.stats.sac.nzhour
    nzmin = tr.stats.sac.nzmin
    nzsec = tr.stats.sac.nzsec
    nzmsec = tr.stats.sac.nzmsec*0.001
    year,month,day = month_day(nzyear,nzjday)
    sac_ref_time = UTCDateTime(year,month,day,nzhour,nzmin,nzsec)+nzmsec
    return sac_ref_time


def get_tr_marker_idx(tr,marker='a'):
    refTime = read_sac_ref_time(tr)
    sp = tr.stats.sampling_rate
    starttime = tr.stats.starttime
    markerTime = refTime+tr.stats.sac[marker]
    idx = int((markerTime - starttime)*sp)
    return idx

def get_tr_data(tr,idx,tb,te):
    """
    tb,te: negative/positive for waveform before/after the marker
    
    """
    delta = tr.stats.delta
    shift1 = int(tb/delta)
    shift2 = int(te/delta)
    data = tr.data[idx+shift1:idx+shift2]
    return data



from numpy.fft import fft,fftshift,fftfreq
import numpy as np
import matplotlib.pyplot as plt


def _convolve(a, b, axis=-1):                                                                                                                                                                                
    return np.apply_along_axis(np.convolve, axis, a, b, mode='same')

def _boundaries(freqs):                                                                                                                                                                                      
    freq_boundaries = np.zeros((len(freqs) + 1,), dtype=float)
    df = freqs[1] - freqs[0]
    freq_boundaries[1:] = freqs + 0.5 * df
    freq_boundaries[0] = freqs[0] - 0.5 * df
    return freq_boundaries

class Bunch(dict):
    """
    A dictionary that also provides access via attributes.

    Additional methods update_values and update_None provide
    control over whether new keys are added to the dictionary
    when updating, and whether an attempt to add a new key is
    ignored or raises a KeyError.

    The Bunch also prints differently than a normal
    dictionary, using str() instead of repr() for its
    keys and values, and in key-sorted order.  The printing
    format can be customized by subclassing with a different
    str_ftm class attribute.  Do not assign directly to this
    class attribute, because that would substitute an instance
    attribute which would then become part of the Bunch, and
    would be reported as such by the keys() method.

    To output a string representation with
    a particular format, without subclassing, use the
    formatted() method.
    """

    str_fmt = "{0!s:<{klen}} : {1!s:>{vlen}}\n"

    def __init__(self, *args, **kwargs):
        """
        *args* can be dictionaries, bunches, or sequences of
        key,value tuples.  *kwargs* can be used to initialize
        or add key, value pairs.
        """
        dict.__init__(self)
        for arg in args:
            self.update(arg)
        self.update(kwargs)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError("'Bunch' object has no attribute '%s'" % name)

    def __setattr__(self, name, value):
        self[name] = value

    def __str__(self):
        return self.formatted()

    def formatted(self, fmt=None, types=False):
        """
        Return a string with keys and/or values or types.

        *fmt* is a format string as used in the str.format() method.

        The str.format() method is called with key, value as positional
        arguments, and klen, vlen as kwargs.  The latter are the maxima
        of the string lengths for the keys and values, respectively,
        up to respective maxima of 20 and 40.
        """
        if fmt is None:
            fmt = self.str_fmt

        items = list(self.items())
        items.sort()

        klens = []
        vlens = []
        for i, (k, v) in enumerate(items):
            lenk = len(str(k))
            if types:
                v = type(v).__name__
            lenv = len(str(v))
            items[i] = (k, v)
            klens.append(lenk)
            vlens.append(lenv)

        klen = min(20, max(klens))
        vlen = min(40, max(vlens))
        slist = [fmt.format(k, v, klen=klen, vlen=vlen) for k, v in items]
        return ''.join(slist)

    def _from_pylines(self, pylines):
        # We can't simply exec the code directly, because in
        # Python 3 the scoping for list comprehensions would
        # lead to a NameError.  Wrapping the code in a function
        # fixes this.
        d = dict()
        lines = ["def _temp_func():\n"]
        lines.extend([f"    {line.rstrip()}\n" for line in pylines])
        lines.extend(["\n    return(locals())\n",
                      "_temp_out = _temp_func()\n",
                      "del(_temp_func)\n"])
        codetext = "".join(lines)
        code = compile(codetext, '<string>', 'exec')
        exec(code, globals(), d)
        self.update(d["_temp_out"])
        return self

    # If I were starting again, I would probably make the following two
    # functions class methods instead of instance methods, so they would
    # follow the factory pattern.  Too late now.

    def from_pyfile(self, filename):
        """
        Read in variables from a python code file.
        """
        with open(filename) as f:
            pylines = f.readlines()
        return self._from_pylines(pylines)

    def from_pystring(self, pystr):
        """
        Read in variables from a python code string.
        """
        pylines = pystr.split('\n')
        return self._from_pylines(pylines)

    def update_values(self, *args, **kw):
        """
        arguments are dictionary-like; if present, they act as
        additional sources of kwargs, with the actual kwargs
        taking precedence.

        One reserved optional kwarg is "strict".  If present and
        True, then any attempt to update with keys that are not
        already in the Bunch instance will raise a KeyError.
        """
        strict = kw.pop("strict", False)
        newkw = dict()
        for d in args:
            newkw.update(d)
        newkw.update(kw)
        self._check_strict(strict, newkw)
        dsub = dict([(k, v) for (k, v) in newkw.items() if k in self])
        self.update(dsub)

    def update_None(self, *args, **kw):
        """
        Similar to update_values, except that an existing value
        will be updated only if it is None.
        """
        strict = kw.pop("strict", False)
        newkw = dict()
        for d in args:
            newkw.update(d)
        newkw.update(kw)
        self._check_strict(strict, newkw)
        dsub = dict([(k, v) for (k, v) in newkw.items()
                                if k in self and self[k] is None])
        self.update(dsub)

    def _check_strict(self, strict, kw):
        if strict:
            bad = set(kw.keys()) - set(self.keys())
            if bad:
                bk = list(bad)
                bk.sort()
                ek = list(self.keys())
                ek.sort()
                raise KeyError(
                    "Update keys %s don't match existing keys %s" % (bk, ek))

def detrend(x, method='linear', axis=-1):
    if method == 'none':
        return x
    if method == 'mean' or method == 'linear':
        try:
            xm = x.mean(axis=axis, keepdims=True)
        except TypeError:
            # older version of numpy
            sh = list(x.shape)
            sh[axis] = 1
            xm = x.mean(axis=axis).reshape(sh)
        y = x - xm
        if method == 'mean':
            return y
        t = np.linspace(-1, 1, x.shape[axis])
        t /= np.sqrt((t ** 2).sum())
        yy = y.swapaxes(axis, -1)
        a = np.dot(yy, t)
        if a.ndim > 0:
            a = a[..., np.newaxis]
        yydetrend = yy - a * t
        return np.ascontiguousarray(yydetrend.swapaxes(-1, axis))
    else:
        raise ValueError("method %s is not recognized" % method)
 
_detrend = detrend # alias to avoid being clobbered by kwarg name                

def window_vals(n, name):
    _name = name
    name = name.lower()
    x = np.arange(n, dtype=float)
    if name == 'boxcar' or name == 'none':
        weights = np.ones((n,), dtype=float)
    elif name == 'triangle':
        weights = 1 - np.abs(x - 0.5 * n) / (0.5 * n)
    elif name == 'welch' or name == 'quadratic':
        weights = 1 - ((x - 0.5 * n) / (0.5 * n)) ** 2
    elif name == 'blackman':
        phi = 2 * np.pi * (x - n / 2) / n
        weights = 0.42 + 0.5 * np.cos(phi) + 0.08 * np.cos(2 * phi)
    elif name == 'hanning':
        phi = 2 * np.pi * x / n
        weights = 0.5 * (1 - np.cos(phi))
    elif name == 'cosine10':
        weights = _Tukey(n, 0.1)
     
    else:
        raise ValueError("name %s is not recognized" % _name)
     
    # Correct for floating point error for all except the boxcar.
    if weights[0] < 1:
        weights[0] = 0
     
    return weights

def _slice_tuple(sl, axis, ndim):
    freqsel = [slice(None)] * ndim
    tup = freqsel[:]
    tup[axis] = sl
    return tuple(tup)


def _welch_params(*args, **kw):                                                                                                                                                                              
    kw = Bunch(kw)
      
    args = [np.array(x, copy=False) for x in np.broadcast_arrays(*args)]
    axis = kw.get('axis', -1)
    npts = args[0].shape[axis]
      
    noverlap = int(kw.overlap * kw.nfft)
      
    weights = window_vals(kw.nfft, kw.window)
      
    step = kw.nfft - noverlap
    ind = np.arange(0, npts - kw.nfft + 1, step)
      
    args += [weights, ind]
    return args


def spectrum(x, y=None, nfft=256, dt=1, detrend='linear',
               window='hanning', overlap=0.5, axis=-1,
               smooth=None):
    """
    Scripts from soest team.
    https://currents.soest.hawaii.edu/ocn_data_analysis/_static/Spectrum.html

    Spectrum and optional cross-spectrum for N-D arrays.

    Rotary spectra are calculated if the inputs are complex.

    detrend can be 'linear' (default), 'mean', 'none', or a function.

    window can be a window function taking nfft as its sole argument,
    or the string name of a numpy window function (e.g. hanning)

    overlap is the fractional overlap, e.g. 0.5 for 50% (default)

    smooth is None or an odd integer. It can be used instead of,
    or in addition to, segment averaging.  To use it exclusively,
    set nfft=None.

    Returns a Bunch with spectrum, frequencies, etc.  The variables
    in the output depend on whether the input is real or complex,
    and on whether an autospectrum or a cross spectrum is being
    calculated.

    """

    if smooth is not None and smooth % 2 != 1:
        raise ValueError("smooth parameter must be None or an odd integer")

    if nfft is None:
        nfft = x.shape[axis]
        nfft -= nfft % 2  # If it was odd, chop off the last point.

    if nfft % 2:
        raise ValueError("nfft must be an even integer")

    # Raw frequencies, negative and positive:
    freqs = fftshift(fftfreq(nfft, dt))  # cycles per unit time used for dt

    if smooth is None:
        n_end = 0
    else:
        n_end = (smooth - 1) // 2  # to be chopped from each end

    kw = dict(nfft=nfft,
              detrend=detrend,
              window=window,
              overlap=overlap,
              axis=axis)
    if y is None:
        x, weights, seg_starts = _welch_params(x, **kw)
    else:
        x, y, weights, seg_starts = _welch_params(x, y, **kw)

    is_complex = (x.dtype.kind == 'c' or y is not None and y.dtype.kind == 'c')

    nsegs = len(seg_starts)

    segshape = list(x.shape)
    segshape[axis] = nfft
    ashape = tuple([nsegs] + segshape)
    fx_k = np.zeros(ashape, np.complex_)
    if y is not None:
        fy_k = fx_k.copy()

    # Make an indexing tuple for the weights.
    bcast = [np.newaxis] * x.ndim
    bcast[axis] = slice(None)
    bcast = tuple(bcast)

    segsel = [slice(None)] * x.ndim

    # Iterate over the segments.  (There might be only one.)
    for i, istart in enumerate(seg_starts):
        indslice = slice(istart, istart + nfft)
        segsel[axis] = indslice
        xseg = x[tuple(segsel)]
        xseg = weights[bcast] * _detrend(xseg, method=detrend, axis=axis)
        fx = fft(xseg, n=nfft, axis=axis)
        fx_k[i] = fx
        if y is not None:
            yseg = y[tuple(segsel)]
            yseg = weights[bcast] * _detrend(yseg, method=detrend, axis=axis)
            fy = fft(yseg, n=nfft, axis=axis)
            fy_k[i] = fy

    fxx = fftshift((np.abs(fx_k) ** 2).mean(axis=0), axes=[axis])
    if y is not None:
        fyy = fftshift((np.abs(fy_k) ** 2).mean(axis=0), axes=[axis])
        fxy = fftshift((np.conj(fx_k) * fy_k).mean(axis=0), axes=[axis])

    # Negative frequencies, excluding Nyquist:
    sl_cw = slice(1, nfft // 2)
    cwtup = _slice_tuple(sl_cw, axis=axis, ndim=x.ndim)

    # Positive frequencies, excluding 0:
    sl_ccw = slice(1 + nfft // 2, None)
    ccwtup = _slice_tuple(sl_ccw, axis=axis, ndim=x.ndim)

    if smooth is not None:
        # start with a boxcar; we can make it more flexible later
        smweights = np.ones((smooth,), dtype=float)
        smweights /= smweights.sum()
        fxx[cwtup] = _convolve(fxx[cwtup], smweights, axis=axis)
        fxx[ccwtup] = _convolve(fxx[ccwtup], smweights, axis=axis)
        if y is not None:
            fyy[cwtup] = _convolve(fyy[cwtup], smweights, axis=axis)
            fyy[ccwtup] = _convolve(fyy[ccwtup], smweights, axis=axis)
            fxy[cwtup] = _convolve(fxy[cwtup], smweights, axis=axis)
            fxy[ccwtup] = _convolve(fxy[ccwtup], smweights, axis=axis)
        # Note: end effects with this algorithm consist of a bias
        # towards zero.  We will adjust the
        # slices to select only the unbiased portions.

    psdnorm = (dt / (weights ** 2).sum())
    psd = fxx * psdnorm
    ps = fxx *  (1.0 / (weights.sum() ** 2))
    if y is not None:
        psd_x = psd
        psd_y = fyy * psdnorm
        psd_xy = fxy * psdnorm
        cohsq = np.abs(fxy) ** 2 / (fxx * fyy)
        cohsq = np.abs(fxy) ** 2 / (np.real(fxx) * np.real(fyy))
        phase = np.angle(fxy)

    if smooth is not None:
        ## Adjust the slices to avoid smoothing end effects:
        # Negative frequencies, excluding Nyquist:
        sl_cw = slice(1 + n_end, nfft // 2 - n_end)
        cwtup = _slice_tuple(sl_cw, axis=axis, ndim=x.ndim)

        # Positive frequencies, excluding 0:
        sl_ccw = slice(1 + nfft // 2 + n_end, -n_end)
        ccwtup = _slice_tuple(sl_ccw, axis=axis, ndim=x.ndim)

    out = Bunch(freqs=freqs[sl_ccw],
                freq_boundaries=_boundaries(freqs[sl_ccw]),
                seg_starts=seg_starts,
                smooth=smooth,
                nfft=nfft,
                detrend=detrend,
                window=window,
                overlap=overlap,
                axis=axis,
                dt = dt
                )

    if not is_complex:
        if y is None:
            out.psd = psd[ccwtup] * 2
            out.ps = ps[ccwtup] * 2
            out.fxx = fxx[ccwtup]
        else:
            out.psd_x = psd_x[ccwtup]
            out.psd_y = psd_y[ccwtup]
            out.psd_xy = psd_xy[ccwtup]
            out.fxx = fxx[ccwtup]
            out.fyy = fyy[ccwtup]
            out.fxy = fxy[ccwtup]
            out.cohsq = cohsq[ccwtup]
            out.phase = phase[ccwtup]

    else:
        out.cwfreqs = -freqs[sl_cw]
        out.cwfreq_boundaries = _boundaries(out.cwfreqs)
        out.ccwfreqs = freqs[sl_ccw]
        out.ccwfreq_boundaries = _boundaries(out.ccwfreqs)

        if y is None:
            out.cwpsd = psd[cwtup]
            out.cwps = ps[cwtup]
            out.ccwpsd = psd[ccwtup]
            out.ccwps = ps[ccwtup]
        else:
            out.cwpsd_x = psd_x[cwtup]
            out.cwpsd_y = psd_y[cwtup]
            out.cwpsd_xy = psd_xy[cwtup]
            out.cwcohsq = cohsq[cwtup]
            out.cwphase = phase[cwtup]

            out.ccwpsd_x = psd_x[ccwtup]
            out.ccwpsd_y = psd_y[ccwtup]
            out.ccwpsd_xy = psd_xy[ccwtup]
            out.ccwcohsq = cohsq[ccwtup]
            out.ccwphase = phase[ccwtup]

    return out

def phase_diff_time(spec,
                    plot=True,
                    x1=None,
                    x2=None,
                    coherence_level=0.95,
                    figsize=(8,8)):
    """
    Calculate phase_diff_time by spectrum result.
    Parameters:
        spec: result of spectrum function
        plot: plot corresponding result figure if True
       x1,x2: data sequences needed for plot
      coherence_level: plot a 95% coherence line in the plot
    """
    weight_cohsq = spec.cohsq**2/(1-spec.cohsq**2)
    W = np.zeros((len(spec.freqs),len(spec.freqs)))
    lenx = len(spec.freqs)
    G = spec.freqs.reshape(lenx,1)
    for i in range(lenx):
        W[i,i] = weight_cohsq[i]
    dobs = spec.phase.reshape(lenx,1)
    dobs = np.rad2deg(dobs)
    WG = np.matmul(W,G)
    Wdobs = np.matmul(W,dobs)
    mest,residuals,rank,sigulars = np.linalg.lstsq(WG,Wdobs,rcond=None)
    deltaT = mest[0][0]*100/360*10   #100Hz/360deg*10ms
    
    if plot:   
        fig,axs = plt.subplots(2,2,figsize=figsize)
        axs = axs.ravel()
        
        # axs[0]
        line1, = axs[0].plot(np.arange(0,len(x1))*spec.dt,x1)
        line2, = axs[0].plot(np.arange(0,len(x2))*spec.dt,x2)
        axs[0].set_xlabel("Time/s")
        axs[0].legend([line1,line2],["trace_1","trace_2"])
        axs[0].set_xlim([0,len(x1)*spec.dt])

        # axs[1]
        line1, = axs[1].plot(spec.freqs,spec.fxx,label="trace_1");
        line2, = axs[1].plot(spec.freqs,spec.fyy,label="trace_2");
        axs[1].legend([line1,line2],["trace_1","trace_2"])
        axs[1].set_xlabel("Frequency/Hz")
        axs[1].set_ylabel("|$S(f)$|")
        axs[1].set_xlim([spec.freqs[0],spec.freqs[-1]])

        # axs[2]
        axs[2].plot(spec.freqs,spec.cohsq)
        axs[2].set_xlabel("Frequency/Hz")
        axs[2].set_ylabel("Coherence")
        axs[2].set_xlim([spec.freqs[0],spec.freqs[-1]])

        # axs[3]
        deg_ms5_pos = spec.freqs*1*spec.dt/2*360
        axs[3].plot(spec.freqs,np.rad2deg(spec.phase))
        line3, = axs[3].plot(spec.freqs,mest[0][0]*spec.freqs,'--')

        line1, = axs[3].plot(spec.freqs,deg_ms5_pos)
        line2, = axs[3].plot(spec.freqs,-deg_ms5_pos)
        axs[3].legend([line1,line2,line3],
                      [f'Reference: +{format(1.0*spec.dt/2*1000,".2f")} ms',\
                       f'Reference: -{format(1.0*spec.dt/2*1000,".2f")} ms',\
                       f'Estimated: {format(deltaT,".2f")} ms'])
        axs[3].set_xlabel("Frequency/Hz")
        axs[3].set_ylabel("Phase/degree")
        axs[3].set_xlim([spec.freqs[0],spec.freqs[-1]])
        plt.tight_layout()
    
    return deltaT



import numpy as np
import os
import re
import logging
from seisloc.sta import load_sta,getNet
from seisloc.utils import gen_wf_files_summary,get_st
from tqdm import tqdm
import obspy
from obspy import Stream
from obspy import UTCDateTime
import multiprocessing as mp

def time_file_list(staDir,prjBtime=None,prjEtime=None):
    '''
    Get the starttime and file path in sequence, append the last
    end time to the time list.
    '''
    dataPths = []
    timePoints = []
    max_endtime = UTCDateTime(1900,1,1)
    for file in os.listdir(staDir):
        file_path = os.path.join(staDir,file)
        assert 'obspy' in globals()
        try:
            st=obspy.read(file_path,headonly=True)
        except:
            continue
        starttimes = []
        endtimes = []
        #----------- Avoid the influence of error times segments ------------    
        if prjBtime!=None:
            for tr in st:
                if tr.stats.starttime < prjBtime:
                    st.remove(tr)
        if prjEtime!=None:
            for tr in st:
                if tr.stats.endtime > prjEtime:
                    st.remove(tr)
        if len(st) == 0:
            continue
        for tr in st:
            starttimes.append(tr.stats.starttime)
            endtimes.append(tr.stats.endtime)
        timePoints.append(min(starttimes))
        if max(endtimes)>max_endtime:
            max_endtime = max(endtimes)
        dataPths.append(file_path)
    tmp = np.array(timePoints)
    dataPths = np.array(dataPths)
    k = tmp.argsort()
    dataPths = dataPths[k]
    timePoints.sort()
    timePoints.append(max_endtime)

    return timePoints,dataPths

def gen_day_split_time_nodes(timePoints,shiftHour):
    """
    Generate time nodes for time split
    """
    tts=timePoints[0] #total start time is the start time of the first file 
    tte=timePoints[-1]#total end time is the end time of the last file
    #start to get the trim point to form day_sac files
    dayTimeNodes=[]
    dayTimeNodes.append(tts)#the first point is the start time of the first file
    #information of the first day
    f_year=timePoints[0].year
    f_month=timePoints[0].month
    f_day=timePoints[0].day
    #the second point should be the start time of the second day
    trim_node=UTCDateTime(f_year,f_month,f_day)+24*60*60+shiftHour
    #if the second point is less than the total end time, add it into list and move to next day
    while trim_node < tte:
        dayTimeNodes.append(trim_node)
        trim_node+=24*60*60
    #append the last time
    dayTimeNodes.append(tte)

    return dayTimeNodes                   

def write_st_to_file(st,outFdr,trimBtime,trimEtime,fileFmt="mseed",staCorrection=False,net=None,sta=None):
    """
    Save traces in the Stream to the desginiated folder in format
    """
    for tr in st:
        if staCorrection==True:
            tr.stats.network = net
            tr.stats.station = sta
        net=tr.stats.network
        sta=tr.stats.station
        chn=tr.stats.channel
        if fileFmt.upper()=="SAC":
            fname=net+"."+sta+"."+chn+"__"+\
                   trimBtime.strftime("%Y%m%dT%H%M%SZ")+"__"+\
                   trimEtime.strftime("%Y%m%dT%H%M%SZ")+".SAC"
        if fileFmt.upper()=="MSEED":
            fname=net+"."+sta+"."+chn+"__"+\
                trimBtime.strftime("%Y%m%dT%H%M%SZ")+"__"+\
                trimEtime.strftime("%Y%m%dT%H%M%SZ")+".mseed"
        tr.write(outFdr+"/"+fname,format=fileFmt)
        logging.info(fname+" writed.")

def load_stream_channels(st):
    chns = []
    for tr in st:
        if tr.stats.channel not in chns:
            chns.append(tr.stats.channel)

    return chns

def get_trim_idxs(trimBtime,trimEtime,timePoints):
    for j in range(len(timePoints)-1):  
        if trimBtime>=timePoints[j] and trimBtime<timePoints[j+1]:
            idxa=j                   
    for k in range(len(timePoints)-1):  
        if trimEtime>=timePoints[k] and trimBtime<timePoints[k+1]:
            idxb=k

    return idxa,idxb

def cut_and_save_day_wf(dataPths,idxa,idxb,trimBtime,trimEtime,fileFmt,outFdr,staCorrection=False,net=None,sta=None):
    #---------------- get channel list -------------------------
    fileChnList = []
    chns = []
    st = Stream()
    while idxa < idxb:
        st+=obspy.read(dataPths[idxa],headonly=True)
        fileChnList.append([dataPths[idxa],st[-1].stats.channel])
        if st[-1].stats.channel not in chns:
            chns.append(st[-1].stats.channel)
        idxa+=1
    del st
    #---------------- process each channel ---------------------
    for chn in chns:
        stUse = Stream()
        for dataPth,channel in fileChnList:
            if channel != chn:
                continue
            stUse += obspy.read(dataPth)
            assert stUse[-1].stats.channel == chn
        if len(stUse)==0:
            continue
        if len(stUse)>1:
            stUse.merge(method=1,fill_value=0)
        stUse.trim(starttime=trimBtime,endtime=trimEtime)
        write_st_to_file(stUse,outFdr,trimBtime,trimEtime,fileFmt=fileFmt,staCorrection=staCorrection,net=net,sta=sta)
        del stUse

def sta_in_dict(sta,staDict):
    flag = 0
    staList = []
    for netName in staDict.keys():
        for staName in staDict[netName].keys():
            if staName == sta:
                staList.append([netName,staName])
    if len(staList)==0:
        flag = 0
    if len(staList)==1:
        flag = 1
    if len(staList)>1:
        flag = 2
        _tmp = ""
        for [netName,staName] in staList:
            _tmp += netName+" "+staName+"; "
        logging.error("func sta_in_dict: more than one station pair: "+_tmp)
    return flag

def check_sta(sta,staDict):
    flag = sta_in_dict(sta,staDict)
    if flag == 0:
        statement=f"{sta} not in station file"
        logging.error(statement)
        raise Exception(statement)
    if flag == 2:
        statement=f"{sta} in more than one network or provided more than once in station file"
        logging.error(statement)
        raise Exception(statement)

def check_station_directories(inDir,staDict):
    stasUse=[]
    for station in os.listdir(inDir): # inDir should only include dirs with name of station
        stationPth = os.path.join(inDir,station)
        if os.path.isdir(stationPth) and station[0] !="\.":
            check_sta(station,staDict)    # check whether sta is in the station list
            stasUse.append(station)
    logging.info("func check_station_...>> Sta qty to be processed: "+str(len(stasUse)))

    return stasUse

def check_trace_status_net_sta(stasUse,staDict,baseDir):
    """
    Check whether the net and sta is correct
    """
    status = True          
    falseRecs = []
    for station in stasUse:
        logging.info("func check_trace_...>> station: "+station)
        stationPth = os.path.join(baseDir,station)
        for item in os.listdir(stationPth):
            itemPth = os.path.join(baseDir,station,item)
            if os.path.isdir(itemPth):
                logging.info("func check_trace_...>> inconsistent folder: "+itemPth)
                continue
            try:
                st = obspy.read(itemPth,headonly=True)
            except:
                logging.info("func check_trace_...>> inconsistent file: "+itemPth)
                continue
            net = st[0].stats.network
            sta = st[0].stats.station
            if net not in staDict:
                logging.info(f"func check_trace_...>> {itemPth}: net [{net}] not in station file")
                falseRecs.append([itemPth,net,sta])
            else:      
                if sta not in staDict[net]:
                    logging.info(f"func check_trace_...>> {itemPth}: sta [{sta}] in station file")
                    falseRecs.append([itemPth,net,sta])
                if sta != station:
                    logging.info(f"func check_trace_...>> {itemPth}: sta [{sta}] != 'dir name' ")
                    falseRecs.append([itemPth,net,sta])

    if len(falseRecs)>0:
        status = False
        if not os.path.exists("Reports"):
            os.mkdir("Reports")
        f = open("Reports/check_trace_status_net_sta.err",'w')
        for rec in falseRecs:
            f.write(f"Error net or sta in folder: {record[0]} net:{record[1]} sta:{record[2]}")
        f.close()
        statement="func check_trace_...>> Error net or sta in trace status, check Reports/check_trace_status_net_sta.err for details"
        logging.info(statement)
        raise Exception(statement)
        
def raw_status_control(rawDir,staFile):
    staDict = load_sta(staFile)
    stasUse = check_station_directories(rawDir,staDict)  # dir name should be sta 
    check_trace_status_net_sta(stasUse,staDict,rawDir)   # trace net and sta

def mp_day_split(inDir,outDir,staFile,fileFmt="mseed",shiftHour=0,parallel=True,processes=10,cutData=True,prjBtime=None,prjEtime=None,staCorrection=False):
    '''
    This function reads in waveform data and split them by days.
    The inDir should be in the strcutre: inDir/staName/dataFiles
    Parameters
        inDir: base folder containing the raw data
       outDir: output directory.
      staFile: station files containing station location information
      fileFmt: "SAC" or "MSEED"
    shiftHour: Shift trimming times to adjust time zones
     parallel: apply multiprocessing
      cutData: if only wish to generate availdays.txt, set cutData to False
     prjBtime: Project begin time. Data should be within the project time span
     prjEtime: Project end time
stacorrection: If True, update net and sta from the station file rather than the internal values of dataset
    '''
    #----------------- Initiation --------------------
    staDict = load_sta(staFile)
    if not os.path.exists(outDir):
        os.mkdir(outDir)
        logging.info(f" Output dirctory {outDir} created")
    #----------------- Clean up information ----------
    for sta in os.listdir(inDir):
        if staCorrection==True:
            net = getNet(sta,staFile)
        staDir = os.path.join(inDir,sta)
        if not os.path.isdir(staDir):
            continue
        timePoints,dataPths = time_file_list(staDir,
                                             prjBtime=prjBtime,
                                             prjEtime=prjEtime)
        if len(dataPths)==0: 
            logging.info(f"{sta} has no available data files")
            continue           
        logging.info(f"{sta} Time span: {timePoints[0]} {timePoints[-1]}")
        #------------- Split the data ----------------
        outFdr = os.path.join(outDir,sta)
        if not os.path.exists(outFdr):
            os.makedirs(outFdr)
        dayTimeNodes = gen_day_split_time_nodes(timePoints,shiftHour)
        if cutData == False: 
            logging.info(f"{sta} cutData is set False, no day split will be conducted!")
            continue                  #only generate availdays.txt
        if parallel:
            pool = mp.Pool(processes=processes)
        for i in range(len(dayTimeNodes)-1):
            trimBtime=dayTimeNodes[i] #trim start time
            trimEtime=dayTimeNodes[i+1]#trim end time
            #get the index of trim_s and trim_e file
            idxa,idxb = get_trim_idxs(trimBtime,trimEtime,timePoints)
            #read in coresponding files
            if parallel:
                pool.apply_async(cut_and_save_day_wf,
                                args=(dataPths,idxa,idxb,trimBtime,trimEtime,fileFmt,outFdr,staCorrection,net,sta))
            else:
                cut_and_save_day_wf(dataPths,idxa,idxb,trimBtime,trimEtime,fileFmt,outFdr,staCorrection,net,sta)
        if parallel:
            pool.close()
            pool.join()
        #gen_wf_files_summary(outFdr)

def cut_single_sta_day_wf(prjBtime,prjEtime,net,sta,rawBase,saveBase,saveFmt,staCorrection=False,fill_value=0):
    wfDir = os.path.join(rawBase,sta)
    saveDir = os.path.join(saveBase,sta)
    if not os.path.exists(saveDir):
        os.mkdir(saveDir)
    loopTime = prjBtime
    while loopTime < prjEtime:
        if staCorrection:
            st = get_st(loopTime,loopTime+24*60*60,wfDir,net=None,sta=None,fill_value=0)
            for tr in st:
                tr.stats.network = net
                tr.stats.station = sta
        else:
            st = get_st(loopTime,loopTime+24*60*60,wfDir,net=net,sta=net,fill_value=0)
        #if len(st)==0:
        #    pass
        st.merge(method=1,fill_value=fill_value)
        for tr in st:
            chn = tr.stats.channel
            if saveFmt.upper()=="SAC":
                fname=net+"."+sta+"."+chn+"_"+\
                       loopTime.strftime("%Y%m%d")+".SAC"
            if saveFmt.upper()=="MSEED":
                fname=net+"."+sta+"."+chn+"_"+\
                       loopTime.strftime("%Y%m%d")+".mseed"
            tr.write(saveDir+"/"+fname,format=saveFmt)
            logging.info(fname+" writed.")
        loopTime += 24*60*60



import matplotlib.pyplot as plt
import re
import numpy as np
from seisloc.utils import draw_vel
from seisloc.sta import load_sta

def cnv_add_evid(vel_in,vel_out="velout.cnv",output_file="velout.cnv.id"):
    """
    The input VELEST cnv file contains event id information which could be 
    processed by VELEST modified by Hardy ZI.
    The output processed by VELEST contains no event id information,
    needs to be addded.
    """
    with open(vel_in,"r") as f:
        vel_in_cont = f.readlines()
    f.close()
    with open(vel_out,"r") as f:
        vel_out_cont = f.readlines()
    f.close()

    f = open(output_file,'w')

    for i,line in enumerate(vel_out_cont):
        if len(vel_out_cont[i])!=1:
            f.write(vel_out_cont[i])
        else:
            f.write(vel_in_cont[i])
    f.close()

def velestmod2ddinv(in_file="velout.mod"):
    """
    Load velocity structure from VELEST *.mod output file to HYPOINVERSE and 
    HYPODD velocity format
    """
    [P_lays,P_vels,S_lays,S_vels] = load_velest_mod(in_file)

    if P_lays != S_lays:
        print("Warning: Layers of P and S are different")
        print(P_lays)
        print(S_lays)
    
    dd_out_file = in_file+".dd"
    f = open(dd_out_file,'w')

    for lay in P_lays:
        f.write(format(lay,"6.3f")+" ")
    f.write("-9\n")

    for vel in P_vels:
        f.write(format(vel,"6.3f")+" ")
    f.write("-9\n")

    for i in range(len(P_vels)):
        f.write(format(P_vels[i]/S_vels[i],'6.3f')+" ")
    f.write("-9\n")
    f.close()

    inv_P_file = in_file+".Pinv"
    f = open(inv_P_file,'w')
    f.write("Velest generated model\n")
    for i in range(len(P_vels)):
        f.write(format(P_vels[i],"5.2f")+" "+format(P_lays[i],'5.2f')+'\n')
    f.close()
    inv_S_file = in_file+".Sinv"
    f = open(inv_S_file,'w')
    f.write("Velest generated model\n")
    for i in range(len(S_vels)):
        f.write(format(S_vels[i],"5.2f")+" "+format(S_lays[i],'5.2f')+'\n')
    f.close()

def load_velest_mod(in_file):
    """
    Read in velest input and output velocity model
    Return an array containing four lists:
    [P_lays,P_vels,S_lays,S_vels]  
    """
    lines = []
    with open(in_file,'r') as f:
        for line in f:
            line = line.strip()
            lines.append(line)
    f.close()

    P_lays = []
    P_vels = []
    S_lays = []
    S_vels = []

    P_lay_qty = int(lines[1].split()[0])
    for line in lines[2:2+P_lay_qty]:
        _vel,_dep,_damp = re.split(" +",line)[:3]
        vel = float(_vel); dep = float(_dep)
        P_vels.append(vel); P_lays.append(dep)

    S_lay_qty = int(lines[2+P_lay_qty])
    if P_lay_qty != S_lay_qty:
        raise Exception("The qty of P and S layers are different")

    for line in lines[3+P_lay_qty:3+P_lay_qty+S_lay_qty]:
        _vel,_dep,_damp = re.split(" +",line)[:3]
        vel = float(_vel); dep = float(_dep)
        S_vels.append(vel); S_lays.append(dep)

    return [P_lays,P_vels,S_lays,S_vels]

class Vel_iter():
    def __init__(self,iter_name):
        self.name = iter_name
        
    def dep_plot(self,ylim=[10,0]):
        """
        plot depth distribution of iteration
        """
        dep = self.deps
        hist,bins = np.histogram(dep,bins = np.arange(0,10,0.5))
        fig,axs = plt.subplots(1,1,figsize=(4,8))
        axs.xaxis.tick_top()
        axs.xaxis.set_label_position("top")
        axs.set_xlabel("Event quantity")
        cnv_bar = axs.barh(bins[:-1]+0.25,hist,height=0.5,color='green',edgecolor='k',alpha=0.5)
        axs.set_ylim(ylim)
        axs.set_ylabel("Depth (km)")
        axs.set_title("Iteration Result",fontsize=16)
        
    def dd_vpvs_format(self,dd_version="1"):
        """
        output hypoDD format velocity lines.
        """
        if dd_version=="2":
            if len(self.P_deps) != len(self.S_deps):
                raise Exception("The P layers qty not equal the S layers qty")
            if self.P_deps != self.S_deps:
                print("Warning: the P-layer and S-layer depths are different, here P layers are used.")
            for dep in self.P_deps:
                print("%6.3f "%dep,end="")
            print("-9\n",end="")
            for vp in self.P_vels:
                print("%6.3f "%vp,end="")
            print("-9\n",end="")
            for i in range(len(self.P_vels)):
                poisson_ratio = self.P_vels[i]/self.S_vels[i]
                print("%6.3f "%poisson_ratio,end="")
            print("-9\n")
        if dd_version=="1":
            for dep in self.P_deps:
                print("%6.3f "%dep,end="")
            print("\n",end="")
            for vp in self.P_vels:
                print("%6.3f "%vp,end="")
            print("\n",end="")

    def gen_inv_mod_files(self):
        inv_P_file = self.name+".Pinv"
        f = open(inv_P_file,'w')
        f.write("Velest generated model\n")
        for i in range(len(self.P_vels)):
            f.write(format(self.P_vels[i],"5.2f")+" "+format(self.P_deps[i],'5.2f')+'\n')
        f.close()
        inv_S_file = self.name+".Sinv"
        f = open(inv_S_file,'w')
        f.write("Velest generated model\n")
        for i in range(len(self.S_vels)):
            f.write(format(self.S_vels[i],"5.2f")+" "+format(self.S_deps[i],'5.2f')+'\n')
        f.close()
        
    def gen_del_files(self,sta_file="/home/zijinping/Desktop/zijinping/resources/stations/sta_sum_202109.txt"):
        """
        Generate station delay files
        """
        sta_dict = load_sta(sta_file)
        iter_name = self.name
        fp = open("P.dly","w")
        fs = open("S.dly","w")
        for sta in self.P_dly.keys():
            # check whether there are stations in different network with the same name
            count = 0
            for tmp_net in sta_dict.keys():
                for tmp_sta in sta_dict[tmp_net].keys():
                    if tmp_sta == sta:
                        count += 1
                        net = tmp_net
            if count > 1:
                print(f"There are several stations with the same station name: {sta}")
                print("The programme couldn't decide the correct network name.")
                print("Please input the correct network name:")
                net = input()
            if count == 0:
                print(f"Station not included in the station file and couldn't get network name")
                net = "  "
            fp.write(f"{format(sta,'5s')} {format(net,'2s')} {format(self.P_dly[sta],'5.2f')}\n")
            fs.write(f"{format(sta,'5s')} {format(net,'2s')} {format(self.S_dly[sta],'5.2f')}\n")
        fp.close()
        fs.close()
                
class Velest():
    """
    The os.getcwd() should be the same with the velest programme working directory 
    """
    def __init__(self,cmn_file = "velest.cmn"):
        self.cmn_file = cmn_file
        self.o_mod = "velout.mod"
        self.load_parameter()
        self.load_i_mod()
        self.iters = []
        self.load_log()
    def __getitem__(self,index):
        return self.iters[index]
    def load_i_mod(self):
        tmp = load_velest_mod(self.i_mod)
        self.i_P_deps = tmp[0]
        self.i_P_vels = tmp[1]
        self.i_S_deps = tmp[2]
        self.i_S_vels = tmp[3]
    def load_parameter(self):
        cont = []
        line_no = 1
        with open(self.cmn_file,'r') as f:
            for line in f:
                line = line.rstrip()
                if len(line)>0 and line[0]=="*":
                    continue
                cont.append(line)
                line_no += 1
        self.i_mod = cont[11]           # input velocity structure
        self.i_sta = cont[12]           # input station file
        self.i_cnv = cont[18]           # input phase file
        
        if len(cont[20].split()) == 0:  # empty line
            self.log_file = "vel.out"
        else:
            self.log_file = cont[20]
        if len(cont[22].split()) == 0:  # empty line
            self.o_cnv = "velout.cnv"
        else:
            self.o_cnv = cont[22]
        if len(cont[23].split()) == 0:  # empty line
            self.o_sta = "velout.cnv"  
        else:
            self.o_sta = cont[23]
            
    def vel_plot(self,iter_name=None,draw_input=True,xlim=[0,8],ylim=[40,0],figsize=(4,6)):
        """
        if iter_name was set, then plot corresponding iteration velocity,
        otherwise the output velocity strucutre will be plotted.
        draw_input: draw the input velocity in dashed line for comparison
        """
        if iter_name == None:
            P_deps,P_vels,S_deps,S_vels = load_velest_mod(self.o_mod)
        else:
            for i,iteration in enumerate(self.iters):
                if iter_name == iteration.name:
                    P_deps = iteration.P_deps
                    P_vels = iteration.P_vels
                    S_deps = iteration.S_deps
                    S_vels = iteration.S_vels
        fig,ax = plt.subplots(1,1,figsize=figsize)
        ax.xaxis.tick_top()
        ax.xaxis.set_label_position("top")
        ax.set_ylabel("Depth (km)",fontsize=16)
        ax.set_xlabel("Velocity (km/s)",fontsize=16)
        ax.set_ylim(ylim)
        if draw_input == True:
            draw_vel(ax,self.i_P_deps,self.i_P_vels,linestyle="--")
            draw_vel(ax,self.i_S_deps,self.i_S_vels,linestyle="--")
        draw_vel(ax,P_deps,P_vels,color='b')
        draw_vel(ax,S_deps,S_vels,color="b")
        plt.show()
        
    def res_plot(self):
        """
        Plot residual of each iteration
        """
        name_list = []
        res_list = []
        for iteration in self.iters:
            name_list.append(iteration.name)
            res_list.append(iteration.residual)
        fig,ax = plt.subplots(1,1,figsize=(6,6))
        x_list = np.arange(0,len(self.iters),1)
        ax.plot(x_list,res_list)
        ax.set_xlabel("Iteration")
        ax.set_ylabel("Residual")
        gap = int(len(x_list)/6)
        plt.xticks(x_list[::gap],name_list[::gap])
        plt.grid()
        plt.show()
        return name_list,res_list

    def rms_plot(self):
        """
        Plot residual of each iteration
        """
        name_list = []
        rms_list = []
        for iteration in self.iters:
            name_list.append(iteration.name)
            rms_list.append(iteration.rms)
        fig,ax = plt.subplots(1,1,figsize=(6,6))
        x_list = np.arange(0,len(self.iters),1)
        ax.plot(x_list,rms_list)
        ax.set_xlabel("Iteration")
        ax.set_ylabel("Residual")
        gap = int(len(x_list)/6)
        plt.xticks(x_list[::gap],name_list[::gap])
        plt.grid()
        plt.show()
        return name_list,rms_list
        
    def load_log(self):
        """
        load information from the log(vel.out) file
        """
        cont = []
        with open(self.log_file,'r') as f:
            for line in f:
                cont.append(line.rstrip())
        f.close()
        i = 0
        while i < len(cont):
            line = cont[i]
            if len(re.split("Iteration",line))>=2:
                line_splits = re.split(" +",line)
                ele_qty = len(line_splits)
                if ele_qty == 5:
                    iter_name = "iter"+line_splits[3]
                if ele_qty == 7:
                    iter_name = "iter"+line_splits[2][:-1]+"B"+line_splits[5]
                new_iter = Vel_iter(iter_name)
                self.iters.append(new_iter)
                residual_line = cont[i+1]
                tmp = residual_line.split()
                new_iter.datavar = float(tmp[1])
                new_iter.residual = float(tmp[5])
                new_iter.rms = float(tmp[8])
                i=i+1
            #-------------------------------------------------------------------------------
            if len(re.split("Velocity model   1",line))>=2:
                j = i+1
                new_iter.P_vels = []
                new_iter.P_deps = []
                while cont[j]!="":
                    _vel,_,_dep = cont[j].split()
                    vel = float(_vel)
                    dep = float(_dep)
                    new_iter.P_vels.append(vel)
                    new_iter.P_deps.append(dep)
                    j=j+1
            #-------------------------------------------------------------------------------
            if len(re.split("Velocity model   2",line))>=2:
                j = i+1
                new_iter.S_vels = []
                new_iter.S_deps = []
                while cont[j]!="":
                    _vel,_,_dep = cont[j].split()[:3]
                    vel = float(_vel)
                    dep = float(_dep)
                    new_iter.S_vels.append(vel)
                    new_iter.S_deps.append(dep)
                    j=j+1 
                i=j
            #-------------------------------------------------------------------------------
            if line == "   stn  ptcor  dpcor":
                j = i+1            # move to the P delay lines
                new_iter.P_dly = {}
                while cont[j]!=" Adjusted station corrections:": # is the start of next part
                    tmp = cont[j].split()
                    sta_qty = len(tmp)/3
                    k=0
                    while k < sta_qty:
                        sta = tmp[k*3+0]
                        delay = float(tmp[k*3+1])
                        new_iter.P_dly[sta]=delay
                        k = k+1
                    j = j+1
            #-------------------------------------------------------------------------------
            if line == "   stn  stcor  dscor":
                j = i+1                            # move to the S delay lines
                new_iter.S_dly = {}
                while cont[j]!="":                 # "" is the end
                    tmp = cont[j].split()          # split values
                    sta_qty = len(tmp)/3           # how many stations
                    k=0                            # denotes the k_th station of line
                    while k < sta_qty:
                        sta = tmp[k*3+0]           # station name
                        delay = float(tmp[k*3+1])  # delay value
                        new_iter.S_dly[sta]=delay  # save value
                        k = k+1                    # move to next station
                    j = j+1                        # move to next line
                i = j                              # reset line indicator
                
            #-------------------------------------------------------------------------------
            if line=="  eq       ot     x      y      z      rms    avres   dot     dx     dy     dz":
                j = i+1
                new_iter.deps=[]
                while cont[j]!="" and cont[j][:5]!="     " :
                    if cont[j][:7]== " ***** ":
                        j = j+1
                        continue
                    tmp = cont[j].split()
                    new_iter.deps.append(float(tmp[4]))
                    j = j + 1
                i = j
            i = i+1
    
    def __len__(self):
        return len(self.iters)




import os
import re
import numpy as np
from tqdm import tqdm
from obspy import UTCDateTime

def load_taup_tvel(tvelFile):
    tvelDict = {}
    f = open(tvelFile,'r')
    lines = f.readlines()
    tvelDict["commentLines"] = lines[:2]
    tvelDict["deps"] = []
    tvelDict["vps"]  = []
    tvelDict["vss"]  = []
    tvelDict["rhos"] = []
    for line in lines[2:]:
        dep,vp,vs,rho = re.split(" +",line.strip())
        tvelDict["deps"].append(float(dep))
        tvelDict["vps"].append(float(vp))
        tvelDict["vss"].append(float(vs))
        tvelDict["rhos"].append(float(rho))
    return tvelDict

def _load_event_dat_etime(timeSeg):
    etime = UTCDateTime.strptime(timeSeg[:8],'%Y%m%d')
    _hr = timeSeg[10:12]
    if _hr == "  ":_hr=" 0"
    _mi = timeSeg[12:14]
    if _mi == "  ":_mi=" 0"
    _sec = timeSeg[14:16]
    if _sec == "  ":_sec=" 0"
    _secf = timeSeg[16:18]
    etime = etime + int(_hr)*60*60 + int(_mi)*60 + int(_sec) + int(_secf)*0.01
    return etime

def load_event_dat(eventDatFile="./event.dat",showExampleLine=False):
    eventDatDict = {}
    f = open(eventDatFile,'r')
    eventLines = f.readlines()
    if showExampleLine:
        print("Head line:"+eventLines[0].rstrip())
    for eventLine in eventLines:
        timeSeg = eventLine[:18]
        etime = _load_event_dat_etime(timeSeg)
        otherSeg = eventLine[18:].strip()
        _lat,_lon,_dep,_mag,_eh,_ez,_rms,_evid = re.split(" +",otherSeg)
        lat,lon,dep,mag,eh,ez,rms = map(float,(_lat,_lon,_dep,_mag,_eh,_ez,_rms))
        evid = int(_evid)
        eventDatDict[evid] = {}
        eventDatDict[evid]['etime'] = etime
        eventDatDict[evid]['evlo'] = lon
        eventDatDict[evid]['evla'] = lat
        eventDatDict[evid]['evdp'] = dep
        eventDatDict[evid]['emag'] = mag
        eventDatDict[evid]['errH'] = eh
        eventDatDict[evid]['errZ'] = ez
        eventDatDict[evid]['rms']  = rms
    return eventDatDict

def write_event_dat(eventDatDict,eventDatFile="event.dat"):
    if os.path.exists(eventDatFile):
        raise Exception("Target eventDatFile existed!")
    f = open(eventDatFile,'w')
    #_estr=
    for evid in eventDatDict.keys():   
        f.write(eventDatDict[evid]['etime'].strftime("%Y%m%d  %H%M%S%f")[:18]+
                format(eventDatDict[evid]['evla'],'>10.4f')+
                format(eventDatDict[evid]['evlo'],'>11.4f')+
                format(eventDatDict[evid]['evdp'],'>11.2f')+
                format(eventDatDict[evid]['emag'],'>7.1f')+
                format(eventDatDict[evid]['errH'],'>8.2f')+
                format(eventDatDict[evid]['errZ'],'>8.2f')+
                format(eventDatDict[evid]['rms'],'>7.2f')+
                format(evid,'>11d')+"\n")

def _write_cnv_file(cnvFile,arc,staChrLen=6):
    f = open(cnvFile,'w')
    for evstr in arc.keys():      # Loop for each event
        evla = arc[evstr]["evla"]
        evlo = arc[evstr]["evlo"]
        evdp = arc[evstr]["evdp"]
        emag = arc[evstr]["emag"]
        etime = arc[evstr]["etime"]
        maxStaAzGap = arc[evstr]['maxStaAzGap']
        rms = arc[evstr]['rms']
        
        _phases = []
        for net,sta,phsType,phsTime,res,weightCode in arc[evstr]["phase"]:
            travTime = phsTime - etime
            phaRecord = sta.ljust(staChrLen," ")+phsType+str(weightCode)+format(travTime,'6.2f')
            _phases.append(phaRecord)
        evid = arc[evstr]["evid"]
        part1 = evstr[2:8]+" "+evstr[8:12]+" "+evstr[12:14]+"."+evstr[14:16]+" "
        part2 = format(evla,"7.4f")+"N"+" "+format(evlo,"8.4f")+"E"+" "
        part3 = format(evdp,"7.2f")+"  "+format(emag,"5.2f")+" "
        part4 = format(maxStaAzGap,'>6d')+" "+format(rms,'9.2f')
        
        f.write(part1+part2+part3+part4)
        i = 0
        for _phase in _phases:
            if i%6==0:
                f.write("\n")
            f.write(_phase)
            i=i+1
        tmp = i%6
        if tmp>0:
            f.write((6-tmp)*(8+staChrLen)*" "+"\n") # add space to fill line
        elif tmp==0:
            f.write("\n")
        if staChrLen==6:
            f.write("    \n")
        else:
            f.write(f"   {str(evid).zfill(6)}\n")
    f.write("9999")              # indicates end of file for VELEST
    f.close()

def write_arc(fileName,arc):
    f = open(fileName,'w')
    for evstr in arc.keys():
        for line in arc[evstr]["lines"]:
            f.write(line+"\n")
    f.close()    

def write_cnv(cnv,cnvFilePth="vel.cnv",staChrLen=6):
    f = open(cnvFilePth,'w')
    for evstr in sorted(cnv.keys()):
        evla = cnv[evstr]["evla"]
        evlo = cnv[evstr]["evlo"]
        evdp = cnv[evstr]["evdp"]
        emag = cnv[evstr]["emag"]
        rms = cnv[evstr]["rms"]
        maxStaAzGap = cnv[evstr]["maxStaAzGap"]
        part1 = evstr[2:8]+" "+evstr[8:12]+" "+evstr[12:14]+"."+evstr[14:16]+" "
        part2 = format(evla,"7.4f")+"N"+" "+format(evlo,"8.4f")+"E"+" "
        part3 = format(evdp,"7.2f")+"  "+format(emag,"5.2f")+" "
        part4 = format(maxStaAzGap,'>6d')+" "+format(rms,'9.2f')
        f.write(part1+part2+part3+part4)
        i=0
        for sta,phsType,travTime,weightCode in cnv[evstr]['phases']:
            _phase=sta.ljust(staChrLen," ")+phsType+str(weightCode)+format(np.round(travTime,2),'6.2f')
            if i%6==0:
                f.write("\n")
            f.write(_phase)
            i=i+1
        tmp = i%6
        if tmp>0:
            f.write((6-tmp)*(8+staChrLen)*" "+"\n") # add space to fill line
        elif tmp==0:
            f.write("\n")
        if staChrLen==6:
            f.write("    \n")
        else:
            #f.write(f"   {str(evid).zfill(6)}\n")
            f.write(f"         \n")
    f.write("9999")              # indicates end of file for VELEST
    f.close()
    
def read_y2000_event_line_mag(line):
    try:
        if len(line) < 126:
            emag = int(line[36:39])*0.01
        else:
            emag = int(line[123:126])*0.01
    except:
        emag = -9
    return emag
    
def read_y2000_event_line(line):
    evstr = line[:16]
    etime = UTCDateTime.strptime(evstr,'%Y%m%d%H%M%S%f')
    evla = int(line[16:18])+int(line[19:23])*0.01/60
    evlo = int(line[23:26])+int(line[27:31])*0.01/60
    evdp = int(line[31:36])*0.01
    emag = read_y2000_event_line_mag(line)
    maxStaAzGap = int(line[42:45])
    rms = int(line[48:52])*0.01
    
    return evstr,etime,evlo,evla,evdp,emag,maxStaAzGap,rms

def read_y2000_phase_line(line):
    sta = re.split(" +",line[:5])[0]
    net = line[5:7]
    phsTimeMinute= UTCDateTime.strptime(line[17:29],"%Y%m%d%H%M")
    if line[14]=="P" and line[47]==" ":
        phsType = "P"
        _secInt = line[29:32]; _secDecimal = line[32:34]
        res = int(line[34:38])*0.01
        weightCode=int(line[16])
    elif line[14]==" " and line[47]=="S":
        phsType = "S"
        _secInt = line[41:44]; _secDecimal = line[44:46]
        res =int(line[50:54])*0.01
        weightCode=int(line[49])
    else:
        raise Exception("Error phase type: line[14] '{line[14]}' and line[47] '{line[47]}'")
        
    if _secInt == "   ": _secInt = "0"
    if _secDecimal == "  ": _secDecimal = "0"
    phsTime = phsTimeMinute+int(_secInt)+int(_secDecimal)*0.01
    
    return sta,net,phsType,phsTime,res,weightCode
    
def load_y2000(y2000File,printLine=False,savePkl=False):
    """
    If printLine is true, each phase line will be printed out
    """
    with open(y2000File,"r") as f1:
        phsLines = f1.readlines()
    arc = {}
    ecount = 0

    print(">>> Loading phases ... ")
    for line in tqdm(phsLines):
        line = line.rstrip()
        if printLine:
            print(line)
        if re.match("\d+",line[:6]):    # line start with digit is summary/event line
            funcReturn = read_y2000_event_line(line)
            evstr,etime,evlo,evla,evdp,emag,maxStaAzGap,rms = funcReturn[:8]
            ecount += 1
            netstas = []
            arc[evstr] = {} 
            arc[evstr]["etime"] = etime
            arc[evstr]["evlo"] = evlo
            arc[evstr]["evla"] = evla
            arc[evstr]["evdp"] = evdp
            arc[evstr]["emag"] = emag
            arc[evstr]["phase"] = []
            arc[evstr]['nsta'] = 0
            arc[evstr]['rms'] = rms
            arc[evstr]['maxStaAzGap']=maxStaAzGap
            arc[evstr]["lines"] = []

        elif line[:6]=='      ':         # The last line
            evid = int(line[66:72])
            arc[evstr]["evid"]=evid            
        else:                            # phase line
            funcReturn = read_y2000_phase_line(line)
            sta,net,phsType,phsTime,res,weightCode = funcReturn[:6]
            travTime = phsTime - etime
            if net+sta not in netstas:
                netstas.append(net+sta)
                arc[evstr]['nsta'] += 1
            arc[evstr]["phase"].append([net,sta,phsType,phsTime,res,weightCode])
        arc[evstr]["lines"].append(line)

    if savePkl:  
        outFile = y2000File+".pkl"
        f = open(out_name,'wb')
        pkl.dump(arc,f)
        f.close()
 
    return arc
    
    
def read_cnv_event_line(line):
    yr = int(line[:2])        
    mo = int(line[2:4])       
    dy = int(line[4:6])   
    hr = int(line[7:9])       
    mi = int(line[9:11])
    secs = float(line[12:17])
    #etime = UTCDateTime(yr+2000,mo,dy,hr,mi,secs)
    etime = UTCDateTime(yr+2000,mo,dy)+\
        hr*60*60+mi*60+secs #modified due to a bug
    evstr = etime.strftime("%Y%m%d%H%M%S%f")[:16]
    evla = float(line[18:25])
    if line[25]=="S":         
        lat = -lat         
    evlo = float(line[27:35])
    if line[35]=="W":         
        lon = -lon         
    evdp = float(line[36:43])
    emag = float(line[45:50])
    if line[54:57] == "   ":
        maxStaAzGap = 0
    else:
        maxStaAzGap=int(line[54:57])
    rms = float(line[63:67])
    
    return evstr,etime,evlo,evla,evdp,emag,maxStaAzGap,rms

def read_cnv_phase_line(line):
    phsRecs = []
    if len(line)%14 == 0:
        segLen = 14
        staChrLen = 6
        nu = int(len(line)/14)
    elif len(line)%13 == 0:
        segLen = 13
        staChrLen = 5
        nu = int(len(line)/13)
    for i in range(nu):
        sta = line[i*segLen:i*segLen+staChrLen].strip()
        phsType = line[i*segLen+staChrLen]
        weightCode = int(line[i*segLen+staChrLen+1])
        travTime = float(line[i*segLen+staChrLen+1+1:i*segLen+staChrLen+1+1+6])
        phsRecs.append([sta,phsType,travTime,weightCode])
    return phsRecs
        

def load_cnv(cnv_file="velout.cnv"):
    """                            
    Load VELEST output file and return a dict with event id as key
    the content is [lon,lat,dep,rms]
    """       
    cnv = {}
    lines = []
    with open(cnv_file,'r') as f:  
        for line in f:
            lines.append(line.rstrip())                
    ecount = 0                  
    for line in lines:
        #print(line)
        if re.match('\d+',line[:2]):  # event line
            if line[:4] == "9999":    # end marker
                continue
            funcReturn = read_cnv_event_line(line)
            evstr,etime,evlo,evla,evdp,emag,maxStaAzGap,rms = funcReturn[:8]
            ecount+=1
            availStas = []
            cnv[evstr] = {}  
            cnv[evstr]["etime"] = etime
            cnv[evstr]["evlo"] = evlo
            cnv[evstr]["evla"] = evla
            cnv[evstr]["evdp"] = evdp
            cnv[evstr]["emag"] = emag
            cnv[evstr]["rms"] = rms
            cnv[evstr]["phases"] = []
            cnv[evstr]['nsta'] = 0 
            cnv[evstr]['maxStaAzGap']=maxStaAzGap
            cnv[evstr]["lines"] = []
        elif len(line)==0:
            cnv[evstr]['evid'] = ecount
        elif line[0]==" ":
            cnv[evstr]['evid'] = int(line)
        elif line[:4] == "9999":
            continue
        else:
            phsRecs = read_cnv_phase_line(line)
            for sta,phsType,travTime,weightCode in phsRecs:
                if sta not in availStas:
                    cnv[evstr]['nsta'] += 1
                    availStas.append(sta)
                cnv[evstr]['phases'].append([sta,phsType,travTime,weightCode])
    return cnv    
    



#----------------------------------------------------------------------------
# coding: utf-8
# Author: ZI,Jinping
# History:
#     2021-04-06 Initial coding
###################################

import os
import numpy as np
from math import radians,cos,acos,sin,asin,sqrt,ceil,pi,floor
import obspy
from obspy import Stream
import glob
import re
from obspy import UTCDateTime
import pandas as pd
import matplotlib.pyplot as plt
import warnings
from obspy.geodetics import gps2dist_azimuth
from tqdm import tqdm
import multiprocessing as mp
import random
import shutil
import subprocess
import time
import pickle
from seisloc.geometry import in_rectangle,loc_by_width
from seisloc.text_io import read_y2000_event_line,read_y2000_phase_line



def invmod2vel(out_file,vp_file,vs_file="",ps_ratio=1.73,vpdamp=1,vsdamp=1,headLine="Velocity model from HYPOINVERSE"):
    """
    Convert hypoinverse velocity model to the velest velocity model

    Parameters:
       out_file: the output velest velocity model file
        vp_file: input hypoinverse P velocity file
        vs_file: input hypoinverse S velocity file, if vs_file == "",
                   the output S velocity will be calculated based on P
                   velocity and ps_ratio
       ps_ratio: used when vs_file==""
         vpdamp: damping factor in the output P velocity model
         vsdamp: damping factor in the output S velocity model
    """
    vp_vels = []
    vp_lays = []
    vs_vels = []
    vs_lays = []
    with open(vp_file,'r') as f:
        cont = f.readlines()
    for i in range(1,len(cont)):                 # Start from the second line
        vp_vel,vp_lay = map(float,cont[i].split())
        vp_vels.append(vp_vel)
        vp_lays.append(vp_lay)
    if vs_file =="":
        vs_lays = vp_lays.copy()
        for vp_vel in vp_vels:
            vs_vels.append(vp_vel/ps_ratio)
    else:
        with open(vs_file,'r') as f:
            cont = f.readlines()
        for i in range(1,len(cont)):             # Start from the second line
            vs_vel,vs_lay = map(float,cont[i].split())
            vs_vels.append(vs_vel)
            vs_lays.append(vs_lay)

    f = open(out_file,'w')
    f.write(headLine+"\n")
    f.write(str(len(vp_vels))+"\n")
    for i in range(len(vp_vels)):
        f.write(format(vp_vels[i],'5.2f'))
        f.write("     ")
        f.write(format(vp_lays[i],'7.2f'))
        f.write("  ")
        f.write(format(vpdamp,'7.3f'))
        f.write("\n")
    f.write(str(len(vs_vels))+"\n")
    for i in range(len(vs_vels)):
        f.write(format(vs_vels[i],'5.2f'))
        f.write("     ")
        f.write(format(vs_lays[i],'7.2f'))
        f.write("  ")
        f.write(format(vpdamp,'7.3f'))
        f.write("\n")

def phs_add_mag(phs_file,mag_file):
    """
    Add magnitude information to the phs file. If no magnitude, set it to -9
    """
    out_file = phs_file+".mag"

    # load in magnitude information
    mag_dict = {}
    with open(mag_file,'r') as f:
        for line in f:
            line = line.strip()
            _evid,_mag = re.split(" +",line)
            evid = int(_evid)
            mag = float(_mag)
            mag_dict[evid] = mag
    
    output_lines = []
    evid_line_idxs = []
    evid_list = []
    i = 0
    tmp_lines = []

    with open(phs_file,'r') as f:
        for line in f:
            line = line.rstrip()
            if line[:5] != "     ":
                tmp_lines.append(line)
            if line[:5] == "     ":
                tmp_lines.append(line)
                evid = int(re.split(" +",line)[1])
                try:
                    e_mag = mag_dict[evid]
                    if e_mag < 0:
                        e_mag = 0
                except:
                    e_mag = 0
                eve_line = tmp_lines[0]
                if len(eve_line) < 126:
                    eve_line = eve_line[:36]+str(int(e_mag*100)).zfill(3)+eve_line[39:]
                output_lines.append(eve_line)  # append the event line
                for line in tmp_lines[1:]:
                    output_lines.append(line)
                tmp_lines = []                 # empty the temporary line
    f.close()

    with open(phs_file+".mag","w") as f:
        for line in output_lines:
            f.write(line+"\n")




def phs_subset(
        phsFile,
        evidLst=[],
        locBdy=[],
        startTime=None,
        endTime=None,
        outFile="",
        reIndex=False,
        startEvid=1
        ):
    """
    subset the *.phs file by event id list or by region location
    evidLst == [] means no evid filter applied
    locBdy == [] means no location filter applied

    Parameters:
    |   phsFile: HYPOINVERSE archieve file
    |   evidLst: selection of events base on evids
    |    locBdy: selction of events according to location boundary
    |   outFile: output file path. Default "" will save as phsFile+".sel"
    |   reIndex: True for re-index events
    | startEvid: startEvid for re-index events
    """
    #----- Preprocessing -----

    evidFltMkr = False
    locFltMkr = False
    
    if len(evidLst) > 0: evidFltMkr = True

    if len(locBdy) > 0:
        locFltMkr = True
        if len(locBdy)!=4:
            raise Exception(
                f"Values qty in locBdy is {len(locBdy)},should be 4."
                )
        loMin,loMax,laMin,laMax = locBdy

    #----- Main content -----
    cont = []
    with open(phsFile,'r') as f:
        for line in f:
            line = line.rstrip()
            cont.append(line)

    out = []
    cache = []
    loopEvid = startEvid
    for line in cont:
        cache.append(line)
        if re.match("\d+",line[:4]):       # Event head line
            lat = int(line[16:18])+int(line[19:23])*0.01/60
            if line[18] == "S":
                lat = -lat
            lon = int(line[23:26])+int(line[27:31])*0.01/60
            if line[26]=="W":
                lon = -lon
            etime = UTCDateTime.strptime(line[:16],"%Y%m%d%H%M%S%f")
        if line[:4]== "    ": # last line of one event
            recordMkr = True
            evid = int(line[66:72])
            if evidFltMkr == True:
                if evid not in evidLst:
                    recordMkr = False
            if locFltMkr == True:
                if lat<laMin or lat>laMax or lon<loMin or lon>loMax:
                    recordMkr = False

            if startTime != None:
                if etime < startTime:
                    recordMkr = False
            if endTime != None:
                if etime > endTime:
                    recordMkr = False

            if recordMkr == True:
                for line in cache[:-1]:
                    out.append(line)
                if reIndex == False:
                    out.append(cache[-1])
                else:
                    out.append(format(loopEvid,'72d'))
                    loopEvid += 1

            cache = []               # reset to empty
    if outFile=='':
        outFile = phsFile+".sel"
    with open(outFile,'w') as f:
        for line in out:
            f.write(line+"\n")
    f.close()

def load_sum_evid(sum_file):
    """
    *.sum file is the catalog summary file after Hyperinverse.
    This function returns a dict:
        -key is event id
        -value is an array with below component:
            --Str format event time "yyyymmddhhmmss**", also the event folder.
            --event longitude
            --event latitude
            --event depth
            --event magnitude
            --event travel time residual
    """
    sum_dict = {}
    with open(sum_file,'r') as f:
        for line in f:
            eve_id=int(line[136:146])
            eve_folder = line[0:16]
            evla = int(line[16:18])+0.01*int(line[19:23])/60
            evlo = int(line[23:26])+0.01*int(line[27:32])/60
            evdp = int(line[31:36])*0.01
            e_mag = int(line[123:126])*0.01
            e_res = int(line[48:52])*0.01
            sum_dict[eve_id] = [eve_folder,evlo,evla,evdp,e_mag,e_res]
    return sum_dict

def load_sum_evstr(sum_file):
    """
    *.sum file is the catalog summary file after Hyperinverse.
    This function returns a dict:
        -key is event time in "yyyymmddhhmmss**" format, same with event folder
        -value is an array with below component:
            --event id
            --event longitude
            --event latitude
            --event depth
            --event magnitude
            --event travel time residual
    """

    sum_dict = {}
    with open(sum_file,'r') as f:
        for line in f:
            eve_id=int(line[136:146])
            eve_folder = line[0:16]
            evla = int(line[16:18])+0.01*int(line[19:23])/60
            evlo = int(line[23:26])+0.01*int(line[27:32])/60
            evdp = int(line[31:36])*0.01
            e_mag = int(line[123:126])*0.01
            e_res = int(line[48:52])*0.01
            sum_dict[eve_folder] = [eve_id,evlo,evla,evdp,e_mag,e_res]
    f.close()
    return sum_dict

def arc_filt(arc_file="Y2000.phs",min_obs=8):
    cont = []
    with open(arc_file,"r") as f:
        cont = f.readlines()
    f.close()
    filt_cont = []
    tmp_cont = []

    for line in cont:
        if line[:4] != "    ":
            tmp_cont.append(line)
        else:
            tmp_cont.append(line)
            if len(tmp_cont) >= min_obs+1+1: # 1 event line, one evid line
                for line in tmp_cont:
                    filt_cont.append(line)
            tmp_cont = []

    with open(arc_file+".filt",'w') as f:
        for line in filt_cont:
            f.write(line)
    f.close()

class Hypoinv():
    def __init__(self,sum_file="out.sum"):
        self.dict_evid = load_sum_evid(sum_file)
        self.dict_evstr = load_sum_evstr(sum_file)
        self.get_locs()
        self.gen_cata_dict()
    
    def get_locs(self):
        self.locs = []
        tmp = 0
        for key in self.dict_evid.keys():
            lon = self.dict_evid[key][1]
            lat = self.dict_evid[key][2]
            dep = self.dict_evid[key][3]
            mag = self.dict_evid[key][4]
            res = self.dict_evid[key][5]
            tmp = tmp+res
            self.locs.append([lon,lat,dep,mag])
        self.locs = np.array(self.locs)
        self.avg_res = tmp/self.locs.shape[0]

    def gen_cata_dict(self):
        self.cataDict = {}
        for key in self.dict_evid.keys():
            evstr = self.dict_evid[key][0]
            etime = UTCDateTime.strptime(evstr,"%Y%m%d%H%M%S%f")
            lon = self.dict_evid[key][1]
            lat = self.dict_evid[key][2]
            dep = self.dict_evid[key][3]
            mag = self.dict_evid[key][4]
            res = self.dict_evid[key][5]
            self.cataDict[key]=[lon,lat,dep,mag,etime]

        
    def crop(self,lonmin,lonmax,latmin,latmax):
        pop_list = []
        for key in self.dict.keys():
            lon,lat,_,_,_ = self.dict[key]
            if lon<lonmin or lon>lonmax or lat<latmin or lat>latmax:
                pop_list.append(key)
                
        for key in pop_list:
            self.dict.pop(key)
    
    def hplot(self,
              xlim=[],
              ylim=[],
              markersize=6,
              size_ratio=1,
              imp_mag=3,
              add_cross=False,
              alonlat=[104,29],
              blonlat=[105,30],
              cross_width=0.1):

        plt.scatter(self.locs[:,0],
                    self.locs[:,1],
                    (self.locs[:,3]+2)*size_ratio,
                    edgecolors = "k",
                    facecolors='none',
                    marker='o',
                    alpha=1)

        kk = np.where(self.locs[:,3]>=imp_mag)
        if len(kk)>0:
            imp = plt.scatter(self.locs[kk,0],
                        self.locs[kk,1],
                        (self.locs[kk,3]+2)*size_ratio*5,
                        edgecolors ='r',
                        facecolors='none',
                        marker='*',
                        alpha=1)
            plt.legend([imp],[f"M$\geq${format(imp_mag,'4.1f')}"])
        if add_cross == True: # draw cross-section plot
            a1lon,a1lat,b1lon,b1lat = loc_by_width(alonlat[0],
                                                   alonlat[1],
                                                   blonlat[0],
                                                   blonlat[1],
                                                   width=cross_width,
                                                   direction="right")
            a2lon,a2lat,b2lon,b2lat = loc_by_width(alonlat[0],
                                                   alonlat[1],
                                                   blonlat[0],
                                                   blonlat[1],
                                                   width=cross_width,
                                                   direction="left")
            plt.plot([a1lon,b1lon,b2lon,a2lon,a1lon],
                     [a1lat,b1lat,b2lat,a2lat,a1lat],
                     linestyle='--',
                     c='darkred')
            plt.plot([alonlat[0],blonlat[0]],[alonlat[1],blonlat[1]],c='darkred')
        if len(xlim) != 0:
            plt.xlim(xlim)
        if len(ylim) != 0: 
            plt.ylim(ylim)
        plt.xlabel("Longitude")
        plt.ylabel("Latitude")
        plt.show()
    
    def vplot(self,alonlat=[],blonlat=[],width=0.1,depmin=0,depmax=10,size_ratio=1,imp_mag=3):
        """
        Description
        """
        length_m,_,_ = gps2dist_azimuth(alonlat[1],alonlat[0],blonlat[1],blonlat[0])
        length_km = length_m/1000
        alon = alonlat[0]; alat = alonlat[1]
        blon = blonlat[0]; blat = blonlat[1]
        results = in_rectangle(self.locs,alon,alat,blon,blat,width)
        jj = np.where(results[:,0]>0)
        plt.scatter(results[jj,1],
                    self.locs[jj,2],
                    marker='o',
                    edgecolors = "k",
                    facecolors='none',
                    s=(self.locs[jj,3]+2)*size_ratio*5)
        tmplocs = self.locs[jj]
        tmpresults = results[jj]
        kk = np.where(tmplocs[:,3]>=imp_mag)

        if len(kk)>0:                 
            imp = plt.scatter(tmpresults[kk,1],
                        tmplocs[kk,2],
                        (tmplocs[kk,3]+2)*size_ratio*10,
                        edgecolors ='red',
                        facecolors='red',
                        marker='*',
                        alpha=1)
            plt.legend([imp],[f"M$\geq${format(imp_mag,'4.1f')}"])
        
        plt.ylim([depmax,depmin])
        plt.xlim([0,length_km])
        plt.xlabel("length (km)")
        plt.ylabel("depth (km)")
        plt.show()

    def depth_hist(self,mag_threshold=-9,depthmin=0,depthmax=10,gap=0.5):
        bins=np.arange(depthmin,depthmax,gap)
        fig,ax = plt.subplots(1,1,figsize=(6,8))
        ax.xaxis.tick_top()
        ax.xaxis.set_label_position("top")
        ax.set_ylabel("Depth (km)",fontsize=16)
        ax.set_xlabel("Event Qty",fontsize=16)
        kk = np.where(self.locs[:,3]>=mag_threshold)
        hist,bins = np.histogram(self.locs[:,2],bins=bins)
        ax.barh(bins[:-1]+gap/2,hist,height=gap,color='gray',edgecolor='k')
        ax.set_ylim([depthmax,depthmin])
        plt.show()
        
    def __repr__(self):
        self.get_locs()
        _qty = f"Hypoinverse catlog with {len(self.locs)} events\n"
        _mag = f"Magnitue range is: {format(np.min(self.locs[:,3]),'4.1f')} to {format(np.max(self.locs[:,3]),'4.1f')}\n"
        _lon = f"Longitude range is: {format(np.min(self.locs[:,0]),'8.3f')} to {format(np.max(self.locs[:,0]),'8.3f')}\n"
        _lat = f"Latitude range is: {format(np.min(self.locs[:,1]),'7.3f')} to {format(np.max(self.locs[:,1]),'7.3f')}\n"
        _dep = f"Depth range is: {format(np.min(self.locs[:,2]),'4.1f')} to {format(np.max(self.locs[:,2]),'4.1f')}\n"
        return _qty+_mag+_lon+_lat+_dep
    
    def __getitem__(self,key):
        if isinstance(key,str):
            return self.dict_evstr[key]
        elif isinstance(key,int):
            return self.dict_evid[key]




import numpy as np
from obspy import Stream, Trace

def lsfit(G,d,W=1,full=False):
    """
    Get the least square fit.
    | G: data kernel
    | d: observations
    | W: weighting matrix
    To be addded: return covariance matrix
    """
    assert len(G.shape)==2
    rows = G.shape[0]
    cols = G.shape[1]
    #-------- set up weight matrix --------
    if isinstance(W,int):
        _W = np.ones(rows)*W
        W = np.diag(_W)
    elif isinstance(W,list) or (isinstance(W,np.ndarray) and len(W.shape)==1):
        W = np.diag(W)
    elif len(W.shape)==2 and W.shape[0] == G.shape[0]:
        pass
    else:
        raise Expception("Wrong W set up!!!")
    WG = np.matmul(W,G)
    GTWG = np.matmul(G.T,WG)
    GTW = np.matmul(G.T,W)
    GTWd = np.matmul(GTW,d)
    GTWG_inv = np.linalg.inv(GTWG)
    m = np.matmul(GTWG_inv,GTWd)

    dpre = np.matmul(WG,m)
    e = dpre - d
    E = np.matmul(e.T,e)
    freedomDeg=len(d) - len(m)
    if freedomDeg == 0: freedomDeg = 1
    sigmad2 = E/freedomDeg

    GTWG_invGTW = np.matmul(GTWG_inv,GTW)

    u,s,vh = np.linalg.svd(GTWG_invGTW)

    s2 = np.matmul(np.diag(s),np.diag(s))
    us2 = np.matmul(u,s2)
    us2uT = np.matmul(us2,u.T)

    if full:
        return m, sigmad2,us2uT*sigmad2
    return m

def weighted_lsfit(G,dobs,W=1,wtMode='log10',wtReciStdMax=5,criteria=0.01,debug=False):
    """ 
    Conduct iterative lsfit until condition fulfilled.
    wtMode: mode for re-weight. 'log10': wt=log10(abs(reciprocal(res))/min(abs(reciprocal)))
    criteria: If ratio of change of m is smaller than the criteria, the iteration will stop
    """
    if isinstance(W,int):
        W = np.eye(G.shape[0],G.shape[0])
    if isinstance(W,np.ndarray) and len(W.shape)==1 and len(W)==len(G):
        W = np.diag(W)
    assert len(W.shape) == 2 and W.shape[0] == W.shape[1]
    
    mnew,sigmad2,covar = lsfit(G,dobs,W,full=True)
    dpre = np.matmul(G,mnew)     # prediction
    dres = np.abs(dobs-dpre)     # residual
    dresReci = 1/dres            # Reciprocal
    if wtMode == "log10":
        Wres = np.diag(W*np.log10(dresReci/np.min(dresReci)))
    count = 1
    if debug: print("Initial weighted residual: ",np.linalg.norm(Wres*dres))
    while True:
        mold = mnew
        mnew,sigmad2New,covarNew = lsfit(G,dobs,Wres,full=True)
        dres = np.abs(dobs-np.matmul(G,mnew))
        if debug:
            print("max/min/norm residual: ",np.max(dres),np.min(dres),np.linalg.norm(dres*Wres))
        dresReci = 1/dres       # Reciprocal
        #>>>>> weighting function >>>>>
        Wres = np.diag(W*np.log10(dresReci/np.min(dresReci)))
        #<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
              
        mdiff = np.abs(mold-mnew)
        moldNorm = np.linalg.norm(mold)
        mdiffNorm = np.linalg.norm(mdiff)
        count += 1
        if mdiffNorm/moldNorm < criteria:
            if debug:
                print(f"Iteration for {count} time timess")
            break
    return mnew

def gaussian_filter(kernel_size, sigma=4, muu=0):
            
    # Initializing value of x,y as grid of kernel size
    # in the range of kernel size
            
    x, y = np.meshgrid(np.linspace(-1, 1, kernel_size),
                       np.linspace(-1, 1, kernel_size))
    dst = np.sqrt(x**2+y**2)
            
    # lower normal part of gaussian
    normal = 1/(2.0 * np.pi * sigma**2)
            
    # Calculating Gaussian filter
    gauss = np.exp(-((dst-muu)**2 / (2.0 * sigma**2))) * normal
    return gauss
            
def avg_filter(kernel_size,dimension=2):
    if dimension==2:
        matrix = np.ones((kernel_size,kernel_size))
    elif dimension==1:
        matrix = np.ones(kernel_size)
    else:
        raise Exception(f"Unaccepted dimension value: {dimension}")
    matrix = matrix/np.sum(matrix)
    return matrix

def mean_period_taup(st,alpha=0.99):
    """
    Calculate mean period of event waveform
    """
    stTaup = st.copy()
    print(len(stTaup))
    for j,trvel in enumerate(st):
        delta = trvel.stats.delta
        tracc = trvel.copy()
        tracc.data[:-1] = (trvel.data[1:]-trvel.data[:-1])/delta
        tracc.data[-1] = 0
        Xi=0
        Di=0
        tauPs = []
        for i in range(len(trvel.data)):
            Xi=alpha*Xi+trvel.data[i]**2
            Di=alpha*Di+tracc.data[i]**2
            tauPi = 2*np.pi*np.sqrt(Xi/Di)
            tauPs.append(tauPi)
        stTaup[j].data = np.array(tauPs)
    
    return stTaup

def interception_two_lines(line1,line2,linspace=100):
    """
    Ouptut the cloest intercepted point on the stright line one w.r.t the line two.
    line1 format: [[x0,y0,...,z0],[x1,y1,...,z1]]
    """
    if not isinstance(line1,np.ndarray):
        line1 = np.array(line1)
    if not isinstance(line2,np.ndarray):
        line2 = np.array(line2)
    assert len(line1.shape) == 2
    assert len(line1.shape) == 2
    lines = np.linspace(line1[0,:],line1[1,:],linspace)

    norms = []
    for i in range(len(lines)):
        A = line2[0,:] - lines[i]
        B = line2[1,:] - lines[i]
        crossAB = np.cross(A,B)
        normAB = np.linalg.norm(crossAB)
        norms.append(normAB)
    k = np.argmin(norms)
    return lines[k]



from seisloc.math import lsfit
import numpy as np
from numpy.linalg import det
from obspy.geodetics import gps2dist_azimuth

def locate_slave(masLon,masLat,masDep,masVels,phaseRecs,model,stas):
    """
    (To be added)
    """
    G = [] # data kernel 
    d = [] # observation 
    W = [] # weights
    # ----- set up kernel -----------
    for i,phaseRec in enumerate(phaseRecs):
        sta,dt,weight,phsType = phaseRec
        staLon = stas.dict["SC"][sta][0]
        staLat = stas.dict["SC"][sta][1]
        dist,az,baz = gps2dist_azimuth(masLat,masLon,staLat,staLon)
        distKm = dist/1000
        distDeg = distKm/111.19
        if phsType.lower() == "p":
            masVel = masVels[0]
        if phsType.lower() == "s":
            masVel = masVels[1]
        phase_list = [phsType.lower(),phsType.upper()]
        arrivals = model.get_ray_paths(source_depth_in_km=masDep,
                                       distance_in_degree=distDeg,
                                       phase_list = phase_list)
        if len(arrivals)>0:
            takeAngle = arrivals[0].takeoff_angle
            phkm = np.sin(np.deg2rad(180-takeAngle))/masVel
            dtdz = -phkm / np.tan(np.deg2rad(takeAngle))
            dtdx = -phkm * np.sin(np.deg2rad(az))
            dtdy = -phkm * np.cos(np.deg2rad(az))
            G.append([dtdx,dtdy,dtdz])
            d.append([dt])
            W.append(weight)
    G = np.array(G)
    # ----- Normalize G columns to enhance the stability of inversion -----
    norm0 = np.linalg.norm(G[:,0])
    norm1 = np.linalg.norm(G[:,1])
    norm2 = np.linalg.norm(G[:,2])
    G[:,0] = G[:,0]/norm0
    G[:,1] = G[:,1]/norm1
    G[:,2] = G[:,2]/norm2
    d = np.array(d)
    W = np.diag(W)
    WG = np.matmul(W,G)
    GTWG = np.matmul(G.T,WG)
    if len(G)<3 or det(GTWG) == 0: # No enough observation or singular matrix
        m = np.zeros((3,1))
    else:
        m = lsfit(G,d,W)
    # ----- G@m=d --> G/norm @ norm*m = d -----
    dxkm = m.ravel()[0]/norm0
    dykm = m.ravel()[1]/norm1
    dzkm = m.ravel()[2]/norm2
    dlon = dxkm/(111.19*np.cos(np.deg2rad(masLat)))
    dlat = dykm/111.19
                            
    return masLon+dlon,masLat+dlat,masDep+dzkm



import os
import obspy
import numpy as np
import time

def SC_tr_remove_spikes(tr):
    """ 
    Remove spikes of the Sichuan continous waveform data
    """
    # more than one zero values
    ks, = np.where(tr.data==0)
    '''
    ascOrder = np.zeros(len(ks))
    desOrder = np.zeros(len(ks))
    ascDiff = ks[1:] - ks[:-1]
    aks, = np.where(ascDiff==1)
    ascOrder[aks] = 1 
    desDiff = ks[:-1] - ks[1:]
    bks, = np.where(desDiff==-1)
    desOrder[bks+1] = 1 
    allOrder = ascOrder.astype('int')|desOrder.astype('int')
    cks = np.where(allOrder==1)
    multiSpikes = ks[cks]
    # only one zero value points
    dks = np.where(allOrder==0)
    singleZeroIdxs = ks[dks]
    ascOrder = np.zeros(len(singleZeroIdxs))
    desOrder = np.zeros(len(singleZeroIdxs))
    ascDelta = tr.data[singleZeroIdxs] - tr.data[singleZeroIdxs-1]
    desDelta = tr.data[singleZeroIdxs+1] - tr.data[singleZeroIdxs]
    multiplyDelta = ascDelta * desDelta
    fks = np.where(np.abs(multiplyDelta)>100000000)
    singleSpikes = singleZeroIdxs[fks]
    print(len(cks[0])+len(fks[0]))
    # update trace
    meanValue = np.sum(tr.data)/(tr.stats.npts - len(multiSpikes)-len(singleSpikes))
    '''
    if len(ks)>0:
        meanValue = np.sum(tr.data)/(tr.stats.npts - len(ks))    
        tr.data -= meanValue
        tr.data[ks] = 0
        tr.detrend("constant")
        tr.detrend("linear")
        tr.data[ks] = 0



import os
import re
import copy
import matplotlib.pyplot as plt
import numpy as np
import matplotlib
from matplotlib.colors import LinearSegmentedColormap
from scipy.cluster.hierarchy import average, fcluster
from scipy.spatial.distance import pdist
from scipy.cluster.hierarchy import dendrogram, linkage
from seisloc.dd import load_DD
from obspy.geodetics import gps2dist_azimuth
import pickle

def load_dtcc(dtcc="dt.cc",save_pkl=False,phases=['P','S'],minobs=4,mean_cc_threshold=0.7):
    """
    If the provided dtcc file is a pkl file, then pickle.load() will be used for loading data
    
    The programme reads in {dtcc} file and calculates average cc value for designated {phases}.
    The average cc value constrain the similarity of two events.
    
    Parameters:
        if save_pkl: a pkl file will be saved with the name {dtcc}.cluster.pkl
                     please note if dtcc file is a .pkl file, even save_pkl is True, it will not 
                     generate a new pkl file.
    """
    if dtcc[-4:] == ".pkl":                         # .pkl file load by pickle
        f = open(dtcc,'rb')
        pair_list = pickle.load(f)
        
    else:                                           # using normal processing
        pair_dict = {}
        i = 0
        f = open(dtcc,'r')
        for line in f:
            line = line.rstrip()
            if line[0] == "#":
                _,_evid1,_evid2,_ = re.split(" +",line)
                evid1 = int(_evid1); evid2 = int(_evid2)
                evid1 = min([evid1,evid2])
                evid2 = max([evid1,evid2])
            else:
                netsta,_diff,_cc,pha = re.split(" +",line)
                try: 
                    pair_dict[evid1][evid2].append(float(_cc))
                except:
                    try:
                        pair_dict[evid1][evid2]=[]
                        pair_dict[evid1][evid2].append(float(_cc))
                    except:
                        try:
                            pair_dict[evid1] = {}
                            pair_dict[evid1][evid2]=[]
                            pair_dict[evid1][evid2].append(float(_cc))
                        except:
                            pass
                            
        pair_list = []
        for key1 in pair_dict.keys():
            for key2 in pair_dict[key1].keys():
                if len(pair_dict[key1][key2])<minobs:
                    continue
                mean_cc = np.mean(pair_dict[key1][key2])
                if  mean_cc < mean_cc_threshold:
                    continue
                pair_list.append([key1,key2,mean_cc])
                
        return pair_list
        
        if save_pkl:                              # generate pkl file to speed loading in future
            dtcc_pkl = dtcc+".clustering.pkl"
            with open(dtcc_pkl,'wb') as f:
                pickle.dump(pair_list,f)

    return pair_list

class Eqcluster():
    def __init__(self,
                 loc_file="hypoDD.reloc",
                 dtcc_file="dt.cc",
                 dtcc_save_pkl=True,
                 dtcc_phases=['P','S'],
                 dtcc_minobs=4,
                 dtcc_mean_cc_threshold=0.7,
                 distWt=0):
        self.cc_threshold = dtcc_mean_cc_threshold
        self.pairs = load_dtcc(dtcc_file,
                               save_pkl=dtcc_save_pkl,
                               phases=dtcc_phases,
                               minobs = dtcc_minobs,
                               mean_cc_threshold=dtcc_mean_cc_threshold)
        self.dd_dict,_ = load_DD(loc_file)

        self.evids = np.array(list(self.dd_dict.keys())) # all event ids in the reloc file
        self.evids.sort()
        self.evids.astype(int)

        self.cc_evids = []                  # evids constrained by cross-correlation
        self.pairs_in = []
        for pair in self.pairs:
            evid1,evid2,cc = pair
            k1 = np.where(self.evids == evid1)
            k2 = np.where(self.evids == evid2)
            if len(k1[0])==1 and len(k2[0])==1:
                self.cc_evids.append(evid1)
                self.cc_evids.append(evid2)
                self.pairs_in.append(pair)
            if len(k1[0])>1 and len(k2[0])>1:
                raise Exception("There are duplicated event ids")
        self.cc_evids = np.array(list(set(self.cc_evids)))
        self.cc_evids.sort()

        # building cc_matrix
        self.cc_matrix = np.ones((len(self.cc_evids),len(self.cc_evids)))
        self.cluster_matrix = np.ones((len(self.cc_evids),len(self.cc_evids)))
        for i in range(self.cc_matrix.shape[0]):
            self.cc_matrix[i,i] = 0
            self.cluster_matrix[i,i] = 0
        if distWt>0:
            for i in range(len(self.cc_evids)):
                evid1 = self.cc_evids[i]
                for j in range(i,len(self.cc_evids)):
                    evid2 = self.cc_evids[j]
                    loEvid1 = self.dd_dict[evid1][0]
                    laEvid1 = self.dd_dict[evid1][1]
                    dpEvid1 = self.dd_dict[evid1][2]
                    loEvid2 = self.dd_dict[evid2][0]
                    laEvid2 = self.dd_dict[evid2][1]
                    dpEvid2 = self.dd_dict[evid2][2]
                    dist12,_,_ = gps2dist_azimuth(laEvid1,loEvid1,laEvid2,loEvid2)
                    dist3D = np.sqrt((dist12/1000)**2 + (dpEvid1-dpEvid2)**2)
                    self.cluster_matrix[i,j] += distWt*dist3D
                    self.cluster_matrix[j,i] += distWt*dist3D
        for pair in self.pairs_in:
            evid1,evid2,cc = pair
            k1 = np.where(self.cc_evids == evid1)
            k2 = np.where(self.cc_evids == evid2)
            self.cc_matrix[k1[0],k2[0]] -= cc
            self.cc_matrix[k2[0],k1[0]] -= cc
            self.cluster_matrix[k1[0],k2[0]] -= cc
            self.cluster_matrix[k2[0],k1[0]] -= cc

        self.in_matrix_evids = self.cc_evids.copy()

    def clustering(self,evids=[],tolerance=0,method='average'):
        """
        Parameters:
        method: please refer to scipy.cluster.hierarchy.linkage
        tolerance: Event with inter-event cc pairs qty lower than tolerance will not be included in clustering
        """
        # build cluster matrix
        if len(evids) == 0:
            self.input_evids = self.evids.copy()
        else:
            self.input_evids = evids
            idxs = []
            for evid in self.input_evids:
                idx = np.where(self.in_matrix_evids==evid)
                idxs.append(idx[0][0])
            self.update_cluster(idxs)

        # apply filter with tolerance
        print(">>> Tolerance filter applied")
        tmp = (self.cluster_matrix < 1)  # True and False matrix
        similar_qty_arr = np.sum(tmp,axis=0)
        kk = np.where(similar_qty_arr>=tolerance)
        self.update_cluster(kk[0])
        print(f"len evids is:{len(self.input_evids)}; ",\
              f"len matrix is: {len(self.in_matrix_evids)};",end=' ')
        print(f"len events not in matrix: {len(self.out_matrix_evids)}")
        
        y = pdist(self.cluster_matrix)
        self.Z = linkage(y,method=method)

    def update_cluster(self,idxs):
        """
        update cluster matrix and cluster evids
        """
        tmp = []
        for idx in idxs:
            tmp.append(self.in_matrix_evids[int(idx)])
        self.in_matrix_evids = tmp
        self.cluster_matrix = self.cluster_matrix[idxs]
        self.cluster_matrix = self.cluster_matrix[:,idxs]
        
        self.out_matrix_evids = []
        for evid in self.input_evids:
            if evid not in self.in_matrix_evids:
                 self.out_matrix_evids.append(evid)
        
        self.in_matrix_evids = np.array(self.in_matrix_evids)
        self.out_matrix_evids = np.array(self.out_matrix_evids)
          
    def dendrogram(self,xlim=[],figsize=(8,6),max_d=None,leaf_rotation=0):
        fig,ax = plt.subplots(1,figsize=figsize)
        if len(xlim)>0:
            ax.set_xlim(xlim)
        self.dn = dendrogram(self.Z,leaf_rotation=leaf_rotation)
        
        if max_d != None:
            plt.axhline(y=max_d, c='k')
            
        plt.xlabel("sample index")
        plt.ylabel("distance")
        
    def fancy_dendrogram(self,
                         truncate_mode='lastp',
                         p=12,
                         show_contracted=True,
                         max_d=None,
                         annotate_above=0,
                         leaf_rotation=0,
                         xytext=(0,-1),
                         no_plot=False):
        """
        https://joernhees.de/blog/2015/08/26/scipy-hierarchical-clustering-and-dendrogram-tutorial/
        """       
        self.dn2 = dendrogram(self.Z,
                             truncate_mode=truncate_mode,
                             p=p,
                             show_contracted=show_contracted,
                             leaf_rotation=leaf_rotation)
        
        if not no_plot:
            plt.title('Hierarchical Clustering Dendrogram (truncated)')
            plt.xlabel("sample index")
            plt.ylabel("distance")
            
        if max_d != None:
            plt.axhline(y=max_d, c='k')
            
            for i,d,c in zip(self.dn2['icoord'],self.dn2['dcoord'],self.dn2['color_list']):
                x = 0.5 * sum(i[1:3])      # "x1, x2, x3, x4", here 'x2,x3'
                y = d[1]                   # "y1, y2, y3, y4", here 'y2'
                if y > annotate_above:
                    plt.plot(x,y,'o',c=c)
                    plt.annotate("%.3g"% y, (x,y),xytext=xytext,
                                textcoords='offset points',
                                va = 'top',
                                ha = 'center')
            
    def heatmap(self,source='cluster',xlim=[],ylim=[],figsize=(5,5),showGroups=True):
        """
        Parameters:
        source: 'dd' to show the orignal time-sequence heatmap
                'cluster' to show the clustered heatmap
        xlim: set plt.xlim if not empty
        ylim: set plt.ylim if not empty
        """
        fig,ax = plt.subplots(1,figsize=figsize)
        if source == "dd":
            disp_matrix = self.cc_matrix
            showGroups = False
            
        elif source == "cluster":            
            self.clustered_matrix = self.cluster_matrix.copy()
            self.clustered_matrix = self.clustered_matrix[self.dn['leaves']]
            self.clustered_matrix = self.clustered_matrix[:,self.dn['leaves']]
            disp_matrix=self.clustered_matrix.copy()
        disp_matrix = 1-disp_matrix   #clustering is conducted by 1-cc; reverse back to normal cc value
        for i in range(disp_matrix.shape[0]):
            disp_matrix[i,i] = 0
        if len(xlim)>0:
            plt.xlim(xlim)
        if len(ylim)>0:
            plt.ylim(ylim)
        self.group_ticks = []
        colors = ["lightgrey",'lightgrey','blue','red']
        nodes = [0,self.cc_threshold,self.cc_threshold,1]
        cmap = LinearSegmentedColormap.from_list("mycmap",list(zip(nodes,colors)))
        plt.pcolormesh(disp_matrix,cmap=cmap,vmin=0,vmax=1)
        if showGroups == True:
            tmpSum=0
            for i in range(self.catQty):
                kks = np.where(self.cluster_category==(i+1))
                print(f"*** Cluster {i+1} has {len(kks[0])} events!")
                tmpSum+=len(kks[0])
                self.group_ticks.append(tmpSum)
                plt.plot([tmpSum,tmpSum],[0,tmpSum],'g-',lw=1)
        plt.gca().set_aspect('equal')
        plt.xlabel("Event No")
        plt.ylabel("Event No")
        plt.colorbar()

    def fcluster(self,max_d,criterion='distance'):
        """
        Categorize clusters and generate self.clusters_evids dict
        refer to: scipy.cluster.hierarchy.fcluster¶
        """
        self.cluster_category = fcluster(self.Z,max_d,criterion=criterion)
        self.catQty = len(set(self.cluster_category))
        print("Num of clusters: ",self.catQty)
        self.clusters_evids = {}

        for cluster_id in np.unique(self.cluster_category):
            self.clusters_evids[cluster_id] = self.get_cluster_evids(cluster_id)

    def mapview(self,clusters=[],xlim=[],ylim=[],figsize=(10,5),mag_base=1):
        """
        Showing corresponding map view of earthquake locations
        Parameters:
        |   clusters: empty list for plot all, otherwide plot clusters in the list
        |       xlim: if not empty, set plt.xlim
        |       ylim: if not empty, set plt.ylim
        |   mag_base: parameters control the magnitude size
        """
        fig,ax = plt.subplots(1,figsize=figsize)

        lons = []
        lats = []
        mags = []
               
        for evid in self.in_matrix_evids:
            lons.append(self.dd_dict[evid][0])
            lats.append(self.dd_dict[evid][1])
            mags.append(self.dd_dict[evid][3])
        self.in_matrix_lons = np.array(lons)
        self.in_matrix_lats = np.array(lats)
        self.in_matrix_mags = np.array(mags)
        cmap = matplotlib.cm.get_cmap('tab20',int(np.max(self.cluster_category)))
        
        if len(clusters)==0:
            plt.scatter(self.in_matrix_lons,self.in_matrix_lats,
                        c=self.cluster_category,
                        s=self.in_matrix_mags-np.min(self.in_matrix_mags)+mag_base,
                        cmap=cmap,vmin=1,
                        vmax=int(np.max(self.cluster_category))+1)
        else:
            kks = np.array([],dtype=int)
            for cid in clusters:
                kk = np.where(self.cluster_category==cid)
                kks = np.concatenate((kks,kk[0]))
            plt.scatter(self.in_matrix_lons[kks],self.in_matrix_lats[kks],
                        c=self.cluster_category[kks],
                        s=self.in_matrix_mags[kks]-np.min(self.in_matrix_mags)+mag_base,
                        cmap=cmap,vmin=1,
                        vmax=int(np.max(self.cluster_category))+1)
                
        if len(xlim)>0:
            plt.xlim(xlim)
        if len(ylim)>0:
            plt.ylim(ylim)
        ax.set_aspect('equal')
        plt.xlabel("Longitude")
        plt.ylabel("Latitude")
        cb = plt.colorbar()
        cb.set_label("Cluster")
        
    def get_cluster_evids(self,cluster_id):
        kk = np.where(self.cluster_category==cluster_id)
        sel_cluster_evids = self.in_matrix_evids[kk[0]]   
        
        return sel_cluster_evids
    
    def __repr__(self):
        str1 = f"Eqcluster object with {len(self.evids)} events in hypoDD relocation file\n"
        str2 = f"{len(self.evids)} events constrained by dt.cc file"
        return str1+str2
    
    def copy(self):
        return copy.deepcopy(self)



#!/usr/bin/evn python
# EQTransformer related functions

import obspy
from obspy import UTCDateTime
import os

def to_mseed(trace,out_folder="./"):
    """
    Save the obspy trace to the format EQTransformer could recognized
    """
    net = trace.stats.network
    sta = trace.stats.station
    chn = trace.stats.channel
    starttime = trace.stats.starttime
    endtime = trace.stats.endtime
    out_name = net+"."+sta+"."+chn+"__"+\
                starttime.strftime("%Y%m%dT%H%M%SZ")+"__"+\
                endtime.strftime("%Y%m%dT%H%M%SZ")+".mseed"\

    trace.write(os.path.join(out_folder,out_name))




import os
import re
import subprocess
from obspy import UTCDateTime
import pandas as pd
from seisloc.sta import load_sta,sta2REAL,getNet

def loadEQTphases(df):
    pdict = {}
    sdict = {}
    for index,rows in df.iterrows():
        ptime = rows["p_arrival_time"]
        pprob = rows["p_probability"]
        stime = rows['s_arrival_time']
        sprob = rows["s_probability"]
        if not pd.isna(ptime):
            ptime = UTCDateTime(ptime)
            pyr = ptime.year
            pmo = ptime.month
            pdy = ptime.day
            psecs = ptime - UTCDateTime(pyr,pmo,pdy)
            pdayStr = ptime.strftime("%Y%m%d")
            if pdayStr not in pdict.keys():
                pdict[pdayStr] = []
            else:
                pdict[pdayStr].append([psecs,pprob])
        if not pd.isna(stime):
            stime = UTCDateTime(stime)
            syr = stime.year
            smo = stime.month
            sdy = stime.day
            ssecs = stime - UTCDateTime(syr,smo,sdy)
            sdayStr = stime.strftime("%Y%m%d")
            if sdayStr not in sdict.keys():
                sdict[sdayStr] = []
            else:
                sdict[sdayStr].append([ssecs,sprob])
    return pdict,sdict

def writeREALpicks(net,sta,pdict,sdict,pickPath):
    """
    pdict,sdict in format of "YYYYMMDD"
    """
    for _day in pdict.keys():
        dayPath = os.path.join(pickPath,_day)
        if not os.path.exists(dayPath):
            os.mkdir(dayPath)
        pFilePath = os.path.join(dayPath,f"{net}.{sta}.P.txt")
        fp = open(pFilePath,'w')
        for record in pdict[_day]:
            psecs = record[0]
            pprob = record[1]
            fp.write(f"{format(psecs,'.3f')} {format(pprob,'4.2f')} 0\n")
        fp.close()
    for _day in sdict.keys():
        dayPath = os.path.join(pickPath,_day)
        if not os.path.exists(dayPath):
            os.mkdir(dayPath)        
        sFilePath = os.path.join(dayPath,f"{net}.{sta}.S.txt")        
        fs = open(sFilePath,'w')        
        for record in sdict[_day]:
            ssecs = record[0]
            sprob = record[1]
            fs.write(f"{format(ssecs,'.3f')} {format(sprob,'4.2f')} 0\n")
        fs.close()

def runREAL(workdir):
    os.chdir(workdir)
    subprocess.run(["perl","runREAL.pl"])

def writeREALpl(assoPath,pickPath,
                R='0.3/20/0.02/1/3',
                G="1.4/20/0.01/1",
                V="6.0/3.3",
                S="5/0/12/1/0.5/0/1.3/1.8",
                workdir="../../Picks/$year$mon$day",
                station="../../sta.real",
                ttime="../../ttdb.txt"):
    """
    Prepare perl scripts for REAL run
    """
    for date in os.listdir(pickPath):
        dayResultPath = os.path.join(assoPath,date)
        if not os.path.exists(dayResultPath):
            os.mkdir(dayResultPath)
        REALplPath = os.path.join(dayResultPath,'runREAL.pl')
        year=date[:4]
        month=date[4:6]
        day=date[6:8]
        with open(os.path.join(REALplPath),"w") as f:
            f.write("#!/usr/bin/perl -w\n")
            f.write(f"$year = \"{year}\";\n")
            f.write(f"$mon = \"{month}\";\n")
            f.write(f"$day = \"{day}\";\n")
            f.write("\n")
            f.write("$D = \"$year/$mon/$day\";\n")
            f.write(f"$R = \"{R}\";\n")
            f.write(f"$G = \"{G}\";\n")
            f.write(f"$V = \"{V}\";\n")
            f.write(f"$S = \"{S}\";\n")
            f.write("\n")
            f.write(f"$dir = \"{workdir}\";\n")
            f.write(f"$station = \"{station}\";\n")
            f.write(f"$ttime = \"{ttime}\";\n")
            f.write("\n")
            f.write("system(\"REAL -D$D -R$R -G$G -S$S -V$V $station $dir $ttime\");\n")
            f.write("print\"REAL -D$D -R$R -G$G -S$S -V$V $station $dir $ttime\";\n")



import os
import glob
import obspy
from obspy import UTCDateTime,read
from obspy.io.sac.sactrace import SACTrace
import shutil
import re
import matplotlib.pyplot as plt
from seisloc.utils import get_st
from seisloc.geometry import spherical_dist
from seisloc.text_io import load_y2000
from seisloc.sta import load_sta
import numpy as np
from tqdm import tqdm
import pickle
import logging
from obspy.taup import TauPyModel

def cut_eve_wf(phs_dict="out.arc",
               sta_dict="sta.txt",
               src_root="day_data",
               tar_root="eve_wf",
               tb=-20,te=60,
               shift_hour=0):
    """
    Parameters:
          phs_dict: HYPOINVERSE output archive file
          sta_dict: station file in format "net sta lon lat ele label"
          src_root: root path for input waveform data
          tar_root: output for event waveform data
                tb: time before event time
                te: time after event time
        shift_hour: shift hours for event time zone, +8 for Beijing time zone
    """
    eve_list = phs_dict.keys()
    print(">>> cut event waveform ... ")
    for eve in tqdm(eve_list):
        str_day = eve[:8]
        #create day folder
        if os.path.exists(os.path.join(tar_root,str_day)):
            pass
        else:
            os.makedirs(os.path.join(tar_root,str_day))
        if os.path.exists(os.path.join(tar_root,str_day,eve)):
            shutil.rmtree(os.path.join(tar_root,str_day,eve))
        os.makedirs(os.path.join(tar_root,str_day,eve))
        e_time = UTCDateTime.strptime(eve,"%Y%m%d%H%M%S%f")-shift_hour*60*60
        starttime = e_time + tb
        endtime = e_time + te
        phases = phs_dict[eve]["phase"]
        processed_sta = []
        for [net,sta,pha,rel_time] in phases:
            src_folder=os.path.join(src_root,sta)
            if sta not in processed_sta:
                processed_sta.append(sta)
                st = get_st(net,sta,starttime,endtime,src_folder)
                if len(st)==0:
                    continue
                for trace in st:
                    chn = trace.stats.channel
                    starttime = trace.stats.starttime
                    b_value = starttime - e_time
                    out_file = os.path.join(tar_root,str_day,eve,f"{sta}.{chn}")
                    trace.write(out_file,format="SAC")
                    sac = SACTrace.read(out_file,headonly=True)
                    net = sac.knetwk
                    sac.lovrok = 1
                    sac.evlo = phs_dict[eve]["eve_loc"][0]
                    sac.evla = phs_dict[eve]["eve_loc"][1]
                    sac.evdp = phs_dict[eve]["eve_loc"][2]
                    sac.stlo = sta_dict[net][sta][0]
                    sac.stla = sta_dict[net][sta][1]
                    sac.stel = sta_dict[net][sta][2]
                    sac.nzyear = e_time.year
                    sac.nzjday = e_time.julday
                    sac.nzhour = e_time.hour
                    sac.nzmin = e_time.minute
                    sac.nzsec = e_time.second
                    sac.nzmsec = e_time.microsecond/1000
                    sac.b = b_value
                    sac.o = 0
                    sac.write(out_file)
                    
def write_eve_wf(phs_dict,tar_root,depth=2):
    '''
    depth = 2 means tar_root/str_day/eve/sta.*
    depth = 1 means tar_root/eve/sta.*
    '''
    eve_list = phs_dict.keys()
    print(">>> write arrival time information ...")
    for eve in tqdm(eve_list):
        str_day = eve[:8]
        phases = phs_dict[eve]["phase"]
        for [net,sta,pha,rel_time] in phases:
            if depth == 2:
                sac_files = glob.glob(os.path.join(tar_root,str_day,eve,sta+"*"))
            if depth == 1:
                sac_files = glob.glob(os.path.join(tar_root,eve,sta+"*"))
 
            if len(sac_files) == 0:
                continue
            if pha == "P":
                for sac_file in sac_files:
                    sac = SACTrace.read(sac_file,headonly=True)
                    sac.a = rel_time
                    logging.info(f">>> write in {eve} {sta} with P arrival {sac.a}")
                    sac.write(sac_file)
            if pha == "S":
                for sac_file in sac_files:
                    sac = SACTrace.read(sac_file,headonly=True)
                    sac.t0 = rel_time
                    logging.info(f">>> write in {eve} {sta} with S arrival {sac.t0}")
                    sac.write(sac_file)

def gen_eve_wf(y2000_file,sta_file,src_root="day_data",tar_root="eve_wf",tb = -20,te = 60,shift_hour=0):
    """
    cut event waveform from continous waveform based on phase file generated by 
    HYPOINVERSE
 
    Parameters:
    y2000_file: phase file, which is the output from HYPOINVERSE
    sta_file: station file in format: net sta lon lat ele
    src_root: source path for continous waveform
    tar_root: target path of eve based waveform
    tb: waveform begin time in relative to the event occurence time
    te: waveform end time in relative to the event ouccurence time
    """
    if not os.path.exists(tar_root):
        os.mkdir(tar_root)
    base_path = os.path.abspath("./")
    sta_dict = load_sta(sta_file)
    phs_dict = load_y2000(y2000_file)
 
    cut_eve_wf(phs_dict,sta_dict,src_root,tar_root,tb,te,shift_hour=shift_hour)
    write_eve_wf(phs_dict,tar_root)

def get_e_time(tr):
    """
    Applied to a sac trace. This function first read sac reference from from sac head.
    Then read o_value from the sac head. Return is the event time
    """
    sac = SACTrace.from_obspy_trace(tr)
    sac_ref_time = sac.reftime    # In UTCDateTime format
    o_value = tr.stats.sac.o 
    event_time = sac_ref_time + o_value     # event origin time
    
    return event_time

def wf_dist_plot(st,
                 xb=0,
                 xe=10,
                 dist_threshold=50,
                 scaling_factor=2,
                 color=None,
                 label_sta=True,
                 out_format="PNG",
                 out_folder="./",
                 figsize=(8,10),
                 align=None,
                 normalize=True):
    '''    
    Description
    ------------
    Plot event waveform by distance. The start time is event origin time.
    Data should be in sac format. The output is a saved file with the title
    of reference time.

    Parameters
    -----------
                st: obspy Stream object
                xb: begin of x in reference to the event time, default 0
                xe: end of x in reference to the event time, default 10
    dist_threshold: set maximum distance for plot
    scaling_factor: The waveform are normalized, increase scaling_facotr to
                    make the waveform plot more obvious
             color: The usage of color is the same as matplotlib.pyplot. 
                    Using default color if not defined.
         label_sta: Whether label station name on the plot.
        out_format: "png","jpg","pdf"..., The same as matplotlib.pyplot.savefig
        out_folder: output folder for the plot file

    Below data information needed:
    |   P arrival: tr.stats.sac.a
    |   S arrival: tr.stats.sac.t0
    |        evla: tr.stats.sac.evla
    |        evlo: tr.stats.sac.evlo
    |        stla: tr.stats.sac.stla
    |        stlo: tr.stats.sac.stlo
    '''
    
    # >>>>>>>>>>>>> Initiation >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    st.detrend("linear")
    st.detrend("constant")
    try:
        e_mag = st[0].stats.sac.mag
    except:
        e_mag = -9
        
    starttime=st[0].stats.starttime
    endtime =st[0].stats.endtime
    sampling_rate = st[0].stats.sampling_rate
    
    event_time = get_e_time(st[0])

    if len(st) == 0:
        print("Warning: Nothing to plot!")

    for tr in st:
        evla = tr.stats.sac.evla
        evlo = tr.stats.sac.evlo
        stla = tr.stats.sac.stla
        stlo = tr.stats.sac.stlo
        #It is recommend to set tr.stats.distance in meters by osbpy guideline
        tr.stats.distance = spherical_dist(evlo,evla,stlo,stla)*111*1000
    
    #Get mininumtime, maximum time, and max distance
    min_time = st[0].stats.starttime
    max_time = st[0].stats.endtime
    max_dist = st[0].stats.distance/1000 # in km
    min_dist = st[0].stats.distance/1000 # in km
    min_dist_idx = 0
    max_dist_idx = 0
    for i,tr in enumerate(st[1:]):
        if tr.stats.starttime < min_time:
            min_time = tr.stats.starttime
        if tr.stats.endtime > max_time:
            max_time = tr.stats.endtime
        if tr.stats.distance/1000 > max_dist:
            max_dist = tr.stats.distance/1000
            max_dist_idx = i
        if tr.stats.distance/1000 < min_dist:
            min_dist = tr.stats.distance/1000
            min_dist_idx = i
    if max_dist>dist_threshold:
        max_dist = dist_threshold
    
    # >>>>>>>>>>>>>>> Plot >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    fig,ax = plt.subplots(1,1,figsize = figsize)
    plt.xlim(xb,xe)
    plt.ylim(0,max_dist*1.1)
    plt.xticks(size=12)
    plt.yticks(size=12)
    plt.xlabel("Time (s)",fontdict={'size':16})
    plt.ylabel("Distance (km)",fontdict={'size':16})
       
    for tr in st:            # Plot trace by trace
        sta = tr.stats.station
        sac = SACTrace.from_obspy_trace(tr)
        tr_ref_time = sac.reftime
        tr_o_value = tr.stats.sac.o
        event_time = tr_ref_time + tr_o_value
        x_start = min_time-event_time
        dist = tr.stats.distance/1000
        
        if dist>max_dist:
            continue
        
        disp_data = tr.data
        if normalize==True:  # Normalize the waveform
            disp_data = tr.data/(max(tr.data) - min(tr.data))
        disp_data = disp_data*scaling_factor
        
        # >>>>>>>>>>>>>> Get align_shift for alignment >>>>>>>>>>>>>>>>>>
        if align == None:
            align_shift = 0
        else:
            align_shift = -tr.stats.sac.__getattribute__(align)

        
        if color == None:
            plt.plot(np.arange(0,len(tr.data))/sampling_rate+x_start+align_shift,
                    disp_data+dist,
                    linewidth = 0.5)
        else:
            plt.plot(np.arange(0,len(tr.data))/sampling_rate+x_start+align_shift,
                    disp_data+dist,
                    color = color,
                    linewidth = 0.5)
        if label_sta:
            plt.text(0.1+xb,dist+0.02,sta,fontsize=12)
        
        #>>>>>>>>>>>>>>>>>>>>> Plot Phase Markers >>>>>>>>>>>>>>>>>>>>>>>>>>>>
        # Plot P arrival if available
        try:
            a = tr.stats.sac.a - tr.stats.sac.o
            gap = 0.5*max_dist/25
            plt.plot([a+align_shift,a+align_shift],[dist-gap,dist+gap],color='b',linewidth=2)
        except:
            pass
        
        # Plot S arrival if available
        try:
            t0 = tr.stats.sac.t0 - tr.stats.sac.o
            gap = 0.5*max_dist/25
            plt.plot([t0+align_shift,t0+align_shift],[dist-gap,dist+gap],color='r',linewidth=2)
        except:
            pass
        # Plot t5 arrival if available
        try:
            t5 = tr.stats.sac.t5 - tr.stats.sac.o
            gap = 0.5*max_dist/25
            plt.plot([t5+align_shift,t5+align_shift],[dist-gap,dist+gap],color='c',linewidth=2)
        except:
            pass
        
        # >>>>> Plot title with magnitude if available >>>>>>>>>>>>>>>>>>>>>>>
        if e_mag != -9:
            plt.title(str(tr_ref_time)+f"_M{e_mag}",fontdict={'size':18})
        else:
            plt.title(str(tr_ref_time),fontdict={'size':18})
    
    # >>>>>>>> Save fig and release memory >>>>>>>>>>>>>>>>>>>>>>
    plt.savefig(os.path.join(out_folder,f"{event_time.strftime('%Y%m%d%H%M%S')}."+out_format.lower()),format=out_format)
    plt.close(fig)



import os
import obspy
from obspy import UTCDateTime,Stream
from obspy.io.sac.sactrace import SACTrace
import re
import matplotlib.pyplot as plt
from seisloc.utils import get_st,extract_set_info
from seisloc.geometry import spherical_dist
from seisloc.sta import load_sta
import numpy as np
from obspy.taup import TauPyModel
from obspy.clients.fdsn import Client
import pandas as pd
import logging
from tqdm import tqdm
from seisloc.noise import xcorr
from numpy import polyfit
import glob
import multiprocessing as mp

def gen_tele_files( datasetPth,
                    staFile,
                    minMag,
                    dist_range,
                    client_name = "IRIS",
                    taup_model="iasp91",
                    tele_dir = "tele_event"):
    """
    Function first searches for suitable tele-event based on condtions provided:
        > starttime
        > endtime
        > minmagnitude
        > distance range in degree[d1, d2]
    Then calculates arrival times for stations and writes into files.
    
    Parameters:
      datasetPth: the path of the dataset. The program will extract information
                    of this dataset, including starttime and endtime
          staFile: station file for station longitude and latitude information
           minMag: minimum magnitude of tele-event for plot
        dist_range: [d1,d2] in degree
       client_name: default "IRIS", check obspy.clients.fdsn.client.Client.__init__()
                     for detail
        taup_model: default "iasp91", check folder obspy/taup/data/ for more model
          tele_dir: target dir

    """    
    logger = logging.getLogger()
    #---------- main program ------------------------------------------------
    if not os.path.exists(tele_dir):
        os.mkdir(tele_dir)
        logger.info("tele dir created.")
    set_info = extract_set_info(datasetPth,staFile)
    c_lon,c_lat = set_info["center"][0]
    starttime = UTCDateTime(set_info['startTime'][:-1])  
    endtime = UTCDateTime(set_info['endTime'][:-1])
    netstas = list(set_info["availYearDays"].keys())
    
    sta_dict = load_sta(staFile)
    client=Client(client_name)
    event_list=client.get_events(starttime=starttime,
                                 endtime=endtime,
                                 minmagnitude=minMag)
    
    columns=["e_id","e_time","e_lon","e_lat","e_dep","e_dist","e_mag","e_mag_type"]
    df = pd.DataFrame(columns=columns)
    for event in event_list:
        e_id = str(event["origins"][0]["resource_id"])[-8:]
        e_time = event["origins"][0]["time"]
        e_lon = event["origins"][0]["longitude"]
        e_lat = event["origins"][0]["latitude"]
        e_dep = event["origins"][0]["depth"]
        e_mag = event['magnitudes'][0]["mag"]
        e_mag_type = event['magnitudes'][0]["magnitude_type"]
        ec_dist = spherical_dist(e_lon,e_lat,c_lon,c_lat) #distance from dataset center
        if ec_dist>=dist_range[0] and ec_dist <= dist_range[1]:
            df.loc[df.shape[0]+1]=[e_id,e_time,e_lon,e_lat,e_dep,ec_dist,e_mag,e_mag_type]

    model = TauPyModel(taup_model)
    for index,row in df.iterrows():
        e_time=row["e_time"]
        e_lon=row["e_lon"]
        e_lat=row["e_lat"]
        e_dep=row["e_dep"]/1000
        e_mag=row["e_mag"]
        e_mag_type=row["e_mag_type"]
        logger.info(f"Now process event: time:{e_time} mag:{e_mag}")
        with open(os.path.join(tele_dir,e_time.strftime("%Y%m%d%H%M%S")+".tele"),"w") as f:
            f.write(f"{e_time} {e_lon} {e_lat} {e_dep} {e_mag} {e_mag_type}\n")
            cont_list=[]
            dist_list=[]
            for netsta in netstas:
                net = netsta[:2]
                sta = netsta[2:]
                sta_lon = sta_dict[net][sta][0]
                sta_lat = sta_dict[net][sta][1]
                es_dist = spherical_dist(e_lon,e_lat,sta_lon,sta_lat)
                arrivals=model.get_travel_times(source_depth_in_km=e_dep,
                                                distance_in_degree=es_dist,
                                                phase_list=["P","S"])
                try:#In case no content error
                    p_arrival=arrivals[0].time
                    s_arrival=arrivals[1].time
                    cont_list.append([netsta,p_arrival,s_arrival,es_dist*111])
                    dist_list.append([es_dist*111])
                    dist_list_sort.append([es_dist*111])
                except:
                    continue
            if len(dist_list) == 0:
                continue
            for dist in sorted(dist_list):
                idx=dist_list.index(dist)
                f.write(f"{cont_list[idx][0]} ")       # netsta       
                f.write(f"{cont_list[idx][1]} ")       # P arrival time
                f.write(f"{cont_list[idx][2]} ")       # S arrival time
                f.write(f"{cont_list[idx][3]}\n")      # event distance
        f.close()


def readTelePhase(cont):
    """
    Read stations tele event arrival time from the tele file
    Parameters:
    |    cont: content list of tele file
    """
    staPhaList=[]              # Array to store the P&S phase time information
    for line in cont[1:]:
        netsta,_P_time,_S_time,_dist=line.split()
        net = netsta[:2]
        sta = netsta[2:]
        P_time = float(_P_time)
        S_time = float(_S_time)
        dist=float(_dist)
        staPhaList.append([netsta,P_time,S_time,dist])
    return staPhaList

def _trimStaTeleWf(net,sta,startTime,endTime,wfFolder,teleWfDir,mode,pad,fill_value):
        if mode == "normal":
            st=get_st(net,sta,startTime,endTime,wfFolder,pad=True,fill_value=0)
        if mode == "SC":
            st=get_st_SC(net,sta,startTime,endTime,wfFolder,pad=True,fll_value=0)
        st = st.select(component="*Z")                # Use Z component
        if len(st) != 0:                                  # len(st)==0 means no waveform
            st[0].write(os.path.join(teleWfDir,net+"_"+sta+".mseed"))  

def trimTeleWf(teleFile,wfRoot,pBefore=50,sAfter=50,mode="normal"):
    """
    test
    Description:
        To plot tele-event waveform, the steps are:
            1. Generate tele-event files
            2. Cut tele-event waveforms of all stations
            3. Make plot of tele-event waveforms
        this functions corresponds to step 2, the waveforms of all stations will
        be saved in one miniseed file under the same tele-file folder with the
        same title but different suffix(".mseed").

        Retrieve from online resources to be developed

    Parameters:
        teleFile: the path of tele-file
          wfRoot: root path for waveform library
         pBefore: start trim point is pBefore seconds before the earliest P
          sAfter: end trim point the pAster seconds after the latest S
            mode: default "normal". "SC" indicates retrieving Sichuan
                    Agency Dataset,uncommon used
    """

    logger = logging.getLogger()
    logger.info(f"Trim tele waveform of tele-event: {teleFile}")
    
    #-------------- load tele file---------------------------------------------
    cont=[]                                 
    with open(teleFile,"r") as f:          
        for line in f:
            cont.append(line.rstrip())
    if len(cont)==1:               # No record, first line is event line
        logger.warn("No station record in tele_file")
        return                              
    
    _etime,_elon,_elat,_edep,_emag,etype = cont[0].split()
    etime=UTCDateTime(_etime[:-1])
    staPhaList = readTelePhase(cont)

    min_P = staPhaList[0][1]
    max_S = staPhaList[-1][2]
                     
    startTime = etime+min_P-pBefore
    endTime = etime+max_S+sAfter

    stsum = Stream()
    for staPhase in tqdm(staPhaList):
        netsta = staPhase[0]
        net = netsta[:2]
        sta = netsta[2:]
     
        wfFolder = os.path.join(wfRoot,sta) # Waveform folder
        if mode == "normal":
            st=get_st(startTime,endTime,wfFolder,net,sta,pad=True,fill_value=0)                                                                                                                            
        if mode == "SC":
            st=get_st_SC(net,sta,startTime,endTime,wfFolder,pad=True,fll_value=0)
        st = st.select(component="*Z")                # Use Z component
        if len(st) != 0:                                  # len(st)==0 means no waveform
            stsum.append(st[0])
    stsum.write(os.path.join(teleFile[:-5]+".mseed"))


def trimTeleWfs(teleRoot="tele_event",wfRoot="day_data",processes=1):
    teleFiles = glob.glob(os.path.join(teleRoot,"*tele"))
    pool = mp.Pool(processes=processes)
    for teleFile in teleFiles:           # Loop for each event
        #trimTeleWf(teleFile,wfRoot)
        pool.apply_async(trimTeleWf,args=(teleFile,wfRoot))
    print(f"Multiprocessing with cores = {processes}")
    pool.close()
    pool.join()

def axisRange(staPhaList,plotPhase,xOffsets,yOffsetRatios):
    """
    Get the axis range for tele waveform plot
    """
    minP = staPhaList[0][1]
    maxP = staPhaList[-1][1]
    minS = staPhaList[0][2]
    maxS = staPhaList[-1][2]
    minDist = staPhaList[0][3]
    maxDist = staPhaList[-1][3]

    ystart = minDist-yOffsetRatios[0]*(maxDist-minDist+0.1)                        
    yend = maxDist+yOffsetRatios[1]*(maxDist-minDist+0.1)

    if plotPhase == "P":
        xstart = minP+xOffsets[0]
        xend = maxP+xOffsets[1]

    elif plotPhase == "S":
        xstart = minS+xOffsets[0]
        xend = maxS+xOffsets[1]

    elif plotPhase == "PS":
        xstart = minP+xOffsets[0]
        xend = maxS+xOffsets[1]
    else:
        raise Exception(f"'plotPhase' parameter {plotPhase} not in ['P','S','PS']")
        
    return xstart,xend,ystart,yend

def plotTeleWf(teleFile,
              wfRoot = "day_data",
              stasSel="all",
              staExclude = [],
              staHighlight=[],
              plotPhase = "P",
              bpRange= [0.5,2],
              xOffsets=[-10,20],
              yOffsetRatios=[0.5,0.5],
              wfNormalize=True,
              wfScaleFactor=1,
              labelStas = "all",
              figsize=(6,8),
              linewidth=1,
              tickHeight=0.05,
              oFormat="pdf",
              saveNameMkr=""):
    """
    Parameters:
          wfRoot: The folder containing wf data in strcture wf_dir/sta_name/'wf files'
         stasSel: list or , station list selected for plot, "all": plot all
     staExclude: stations excluded in the plot,used for problem staion
   staHighlight: station waveform to be highlighted will be drawn in green
      plotPhase: "P","S" or "PS". "P" only plot P arrival, "S"  only plot S arrival.
                    "PS" means both P arrival and S arrival will be presented
      labelStas: False means no label, empty list means all, else label station in list
    """
    logger = logging.getLogger()
    logger.info("plot tele event: "+teleFile)

    mseed_file = teleFile.rsplit(".",1)[0]+".mseed"
    if not os.path.exists(mseed_file):
        logging.error("mseed file not exits, did you run the trimTeleWf?")
        return                              # end the programme
    else:
        saved_st = obspy.read(mseed_file)
    #-------------- load tele file---------------------------------------------
    cont=[]                                 # Store content from file
    with open(teleFile,"r") as f:          # Load starts
        for line in f:
            cont.append(line.rstrip())
    f.close()                               # Load finishes
    if len(cont)==1:
        logger.warn("No station record in tele_file")
        return                              # No record, first line is event line
    _etime,_elon,_elat,_edep,_emag,etype = cont[0].split()
    etime=UTCDateTime(_etime[:-1])
    elon=float(_elon)
    elat=float(_elat)
    emag=float(_emag)

    staPhaList = readTelePhase(cont)

    xstart,xend,ystart,yend = axisRange(staPhaList,plotPhase,xOffsets,yOffsetRatios)

    plt.close()                               # close previous figure as this is inside loop 
    fig,ax = plt.subplots(1,1,figsize=figsize) # Initiate a figure with size 8x10 inch        

    plt.axis([xstart,xend,ystart,yend]) # Set axis
    tele = re.split("/",teleFile)[-1]
    title = etime.strftime("%Y-%m-%d %H:%M:%S")
    plt.title(f'Tele Event {title} M{emag}')           # Set the title
    plt.xlabel("Time (s)")
    plt.ylabel("Distance (km)")
    # Draw event waveform, P and S arrival markers for each station
    for staPhase in tqdm(staPhaList):
        netsta = staPhase[0]
        net = netsta[:2]
        sta = netsta[2:]

        if stasSel != "all": # Then stasSel should be a list
            if sta not in stasSel:
                continue
        if sta in staExclude:
            continue

        P_time = staPhase[1]
        S_time = staPhase[2]
        dist = staPhase[3]
        st = saved_st.select(network=net,station=sta)
        st.trim(starttime = etime+xstart,
                    endtime = etime+xend,
                    pad = True)

        if len(st) != 0:                                  # len(st)==0 means no waveform
            st = st.select(component="*Z")                # Use Z component
            sampling_rate = st[0].stats.sampling_rate
            chn = st[0].stats.channel
            st[0].detrend("linear")                       # Remove linear trend
            st[0].detrend("constant")                     # Remove mean
            st.filter("bandpass",freqmin=bpRange[0],freqmax=bpRange[1],corners=2)
            if wfNormalize:
                data = st[0].data.copy()
                if max(data) != min(data):
                    st[0].data=data/(max(data) - min(data)) # Normalize data
                st[0].data = st[0].data * wfScaleFactor
            # Draw waveform
            if sta in staHighlight:                               # color='k' means black
                plt.plot(np.arange(0,len(st[0].data))*1/sampling_rate+xStart,
                         st[0].data+dist,
                         color='darkred',
                         linewidth=3*linewidth)
            else:                                         # color='g' means green
                plt.plot(np.arange(0,len(st[0].data))*1/sampling_rate+xstart,
                         st[0].data+dist,
                         color='k',
                         linewidth=linewidth)
            # Plot P arrival marker in red
            P_marker, = plt.plot([P_time,P_time],[dist-tickHeight,dist+tickHeight],color='r',linewidth=2)
            # Plot S arrival marker in blue
            S_marker, = plt.plot([S_time,S_time],[dist-tickHeight,dist+tickHeight],color='b',linewidth=2)
            if labelStas=="all":
                plt.text(xstart,dist,f'{sta}',color='darkred',fontsize=12)
            elif sta in labelStas:
                plt.text(xstart,dist,f'{sta}',color='darkred',fontsize=12)
    try:
        plt.legend([P_marker,S_marker],['tele P','tele S'],loc='upper right')
    except:
        pass
    plt.tight_layout()
    if oFormat.lower()=="pdf":
        plt.savefig(teleFile[:-5]+saveNameMkr+".pdf")
    if oFormat.lower()=="jpg" or oFormat.lower=="jpeg":
        plt.savefig(teleFile[:-5]+saveNameMkr+".jpg")
    if oFormat.lower()=="png":
        plt.savefig(teleFile[:-5]+saveNameMkr+".png")

def plotTeleWfs(teleRoot="tele_event",wfRoot="day_data",
                          stasSel='all',
                          staExclude = [],
                          plotPhase = "P",
                          bpRange= [0.5,2],
                          xOffsets=[-10,20],
                          yOffsetRatios=[0.01,0.05],
                          wfNormalize=True,
                          wfScaleFactor=0.06,
                          labelStas = "all",
                          figsize=(8,12),
                          linewidth=0.5,
                          tickHeight=0.01,
                          oFormat="png",
                          saveNameMkr=""):
    teleFiles = glob.glob(os.path.join(teleRoot,"*tele"))
    for teleFile in teleFiles:           # Loop for each event
            plotTeleWf(teleFile,
                          wfRoot = wfRoot,
                          stasSel=stasSel,
                          staExclude = staExclude,
                          plotPhase = plotPhase,
                          bpRange= bpRange,
                          xOffsets=xOffsets,
                          yOffsetRatios=yOffsetRatios,
                          wfNormalize=wfNormalize,
                          wfScaleFactor=wfScaleFactor,
                          labelStas = labelStas,
                          figsize=figsize,
                          linewidth=linewidth,
                          tickHeight=tickHeight,
                          oFormat=oFormat,
                          saveNameMkr=saveNameMkr)


def plotTeleDiffs(plotPhase="P",
                 tb=-5,
                 te=20,
                 root="tele_event",
                 stasSel="all",
                 maxlag=100,
                 freqRange=[0.5,2],
                 decimateFactor=5,
                 threshold = 0.5,
                 figsize=(10,6),
                 saveNameMkr=""):
    """
    This function reads in the corresponding tele event minseed files,
    do cross-correlation to find large shift stations and make plot.
    
    Parameter:
      maxlag: maxlag data points. The total calculation times is 2*maxlag+1
    """
    logger = logging.getLogger()
#---------------------------------------------------------------------------    
    tele_files = glob.glob(os.path.join(root,"*tele"))
    for tele_file in tqdm(tele_files):
        with open(tele_file,'r') as f:
            lines = f.readlines()
        f.close()
        if not os.path.exists(tele_file[:-4]+"mseed"):
            print(tele_file[:-4]+"mseed "+"not existed!")
            continue
        st = obspy.read(tele_file[:-4]+"mseed")
        st.detrend("linear")
        st.detrend("constant")
        st.taper(max_percentage=0.05)
        if stasSel!="all":
            assert isinstance(stasSel,list)
            st_new = Stream()
            for tr in st:
                if tr.stats.station in stasSel:
                    st_new.append(tr)
        else:
            st_new = st
        st_new = st_new.decimate(factor=decimateFactor)
        st_new.filter("bandpass",freqmin=freqRange[0],freqmax=freqRange[1],zerophase=True)

        ref_sta_set =False

        maxtimes = []
        sta_sequence = []
        if len(st_new) == 0: # debug
            continue
        delta = st_new[0].stats.delta
        
        str_time,_,_,_dep,_mag,type = lines[0].split()
        etime = UTCDateTime(str_time)
        
        for line in lines[1:]:
            line = line.rstrip()
            netsta,_p,_s,_dist = line.split()
            net = netsta[:2]
            sta = netsta[2:]
            st_sel = st_new.select(network=net,station=sta,component='*Z')
            if len(st_sel)==0:
                continue
            if plotPhase == "P":
                ttb = etime + float(_p) + tb
                tte = etime + float(_p)+ te
            if plotPhase == "S":
                ttb = etime + float(_s) + tb
                tte = etime + float(_s)+ te
            if plotPhase == "PS":
                ttb = etime + float(_p) + tb
                tte = etime + float(_s)+ te
            st_sel = st_sel.trim(starttime=ttb,endtime=tte)
            sta_tr = st_sel[0]
            if not ref_sta_set:
                ref_sta = sta
                ref_sta_set = True
                continue
            ref_tr = st_new.select(station=ref_sta,component="*Z")[0]
            try:
                corr_result = xcorr(sta_tr.data,ref_tr.data,maxlag)
            except:
                continue
            corr_result = list(corr_result/(np.linalg.norm(sta_tr.data)*np.linalg.norm(ref_tr.data)))
            max_corr = max(corr_result)
            if max_corr < 0.7:
                continue
            max_index = corr_result.index(max_corr)
            maxtimes.append((max_index-maxlag)*delta)
            sta_sequence.append(sta)

        fig,ax = plt.subplots(1,1,figsize=figsize)
        try:
            p = polyfit(np.arange(len(maxtimes)),maxtimes,deg=1)
        except:
            continue
        x = np.linspace(0,len(maxtimes)-1,len(maxtimes))
        y = x*p[0]+p[1]
        diffs = maxtimes-y
        plt.scatter(x,diffs,c=np.abs(diffs),s=40,edgecolor='k',cmap="rainbow",vmin=0.2,vmax=4)
        #plt.ylim([-1,5])
        plt.xlim([0,len(maxtimes)])
        plt.ylabel("Time (s)")

        for i in range(len(sta_sequence)):
            if np.abs(diffs[i])>threshold:
                plt.text(i,diffs[i],sta_sequence[i])
        plt.title(f"Tele Event {etime} M{_mag}")
        plt.savefig(tele_file[:-5]+saveNameMkr+".jpg")



import time
import functools

def timer(func):
    functools.wraps(func)
    def wrapper_timer(*args,**kwargs):
        startTime = time.perf_counter()
        func(*args,**kwargs)
        endTime = time.perf_counter()
        runTime = endTime - startTime
        print(f"Finished {func.__name__!r} in {runTime:.4f} secs")
    return wrapper_timer



import matplotlib.pyplot as plt
import numpy as np

def draw_vel(dep_list,vel_list,ax=None,color='k',linestyle='-',label=""):
    """
    Draw velocity line on the ax based on the depth list and velocity list
    """
    points_list = []
    points_list.append([dep_list[0],vel_list[0]])
    for i in range(1,len(dep_list)):
        points_list.append([dep_list[i],vel_list[i-1]])
        points_list.append([dep_list[i],vel_list[i]])
        
    points_list = np.array(points_list)
    if ax==None:
        ax = plt.gca()
    line, = ax.plot(points_list[:,1],points_list[:,0],color=color,linestyle=linestyle,label=label)
    return line
